{{
Object file:    Db_FullDuplexSerial.spin
Version:        1.2.1*
Date:           2006 - 2011
Author:         Chip Gracey, Jeff Martin, Daniel Harris, Tom Dinger
Company:        Parallax Semiconductor
Email:          dharris@parallaxsemiconductor.com
                propeller@tomdinger.net
Licensing:      MIT License - see end of file for terms of use.

Description:
This driver, once started, implements a serial port in one cog.

Revision History:
       -02/23/2024 [TD] Merged the "best" of the _rr004 version with the -wCTS
                   version; other changes:
                   - Separate tx and rx buffer pointers and sizes in the
                     parameter block
                   - Removed the power-of-two requirement for buffer sizes.
                     (the PASM code doesn't care)
                   - removed the Propeller font extended characters, to make
                     the file ASCII-friendly.
                   - This file contains just the Spin functions for normal
                     operations once the PASM cog has started. The PASM code
                     is now in a separate file.
                   TODO: This is now different enough, perhaps I should change
                   the version numbering...

       -12/03/2015 Changed "ifnot ctsPinSet" to "if !ctsPinSet" to correct
                   erroneous use of pin 0 as CTS pin when not explicitly set
       - 2/14/2015 Moved ctsPinSet to DAT with initial zero value
       - 2/26/2014 Modified to permit optional single sided flow control
                  (transmit contingent on CTS high)
       - 1/20/2013 Modified to permit different buffer sizes by incorporating
                    code from Parallax Serial Terminal.
v1.2.1 - 5/1/2011 Added extra comments and demonstration code to bring up
                   to gold standard.
v1.2 - 5/7/2009 Fixed bug in dec method causing largest negative value
                (-2,147,483,648) to be output as -0.
v1.1 - 3/1/2006 First official release.

 
=============================================
        Connection Diagram
=============================================

        +---------+
        |         |         
        |    rxPin|---< TTL level RX line
        |    txPin|---> TTL level TX line
        |   ctsPin|---< CTS line (optional)
        |         |         
        +---------+
         Propeller
            MCU
          (P8X32A)

Components:
N/A

=============================================                  
}}
CON
  NO_CTS        = $1E0 '[MMO]   ' [TD] Must have a value > 32 to indicate "not used"

VAR

  ' Object global variable declarations

  long  cog           'cog flag/id  -- CogID + 1

  '13 words, MUST be contiguous
  word  rx_head                 
  word  rx_tail
  word  tx_head
  word  tx_tail
  word  rx_pin
  word  tx_pin
  word  rxtx_mode
  word  bit_ticks
  word  rx_buffer_ptr
  word  tx_buffer_ptr
  word  rx_bufsiz
  word  tx_bufsiz
  word  cts_pin

'=========================================================================
PUB Stop
{{
   Stop serial driver if it has already been started - frees the cog

   Parameters: none
   return:     none

   example usage: serial.Stop

   expected outcome of example usage call: Stops an already started serial port.
}}
  if cog
    cogstop(cog~ - 1)                   'if the driver is already running, stop the cog
  wordfill(@rx_head, 0, 13)             'zero out configuration variables

'=========================================================================
PUB GetInitVars
{{
  Returns the address of the block of configuration variables, that will be
  initialized when the PASM cog is loaded.

  Parameters: none
  return:     Address of the first initialization variable -- see above.

  example use: serial_init_vars := serial.GetInitVars

  expected outcome of example use call: The variable serial_init_vars gets the
                                        address of the 13 WORDs for the serial
                                        configuration.
}}
  return @rx_head

'=========================================================================
PUB FinishInit(cid, rx_bufptr, rx_size, tx_bufptr, tx_size)
{{
   Complete initialization, using the cog ID and buffer information from the
   init of the PASM cog.

   Parameters: cid = cog ID + 1 of the cog running the PASM serial driver
               rx_bufptr -- hub address of a byte buffer of size rx_size
               rx_size -- size in bytes of the receive buffer (rx_bufptr)
               tx_bufptr -- hub address of a byte buffer of size tx_size
               tx_size -- size in bytes the of transmit buffer (tx_bufptr)
   return:     none

   example use: serial.FinishInit(CogIDplus1, @RxBuffer, RX_BUFSIZE, @TxBuffer, TX_BUFSIZE)

   expected outcome of example use call: The serial services are running and can be
                                         used to send and receive bytes.
}}
  cog := cid
  tx_buffer_ptr := tx_bufptr
  rx_bufsiz     := rx_size
  tx_bufsiz     := tx_size
  ' This must be done last, because it will unblock the PASM cog, which will
  ' immediately read the rest of the parameters to finish its initialization.
  rx_buffer_ptr := rx_bufptr

'=========================================================================
PUB RxFlush
{{
   Continuously pops the head of the receive buffer until no bytes remain.
   
   Parameters: none
   return:     none

   example usage: serial.RxFlush

   expected outcome of example usage call: Receive bffer will be cleared.
}}

  repeat while RxCheck => 0                             'Call RxCheck until buffer is empty

'=========================================================================
PUB RxAvail : truefalse
{{
   Returns true (-1) if any bytes are available
   
   Parameters: none
   return:     false (0) if no bytes available; true (-1) if bytes available

   example usage: serial.RxAvail

   expected outcome of example usage call: Returns a true/false value for bytes
                                           available, but no other changes.
}}

  truefalse := rx_tail <> rx_head
  
'=========================================================================
PUB RxCheck : rxByte
{{
   Check if a byte is waiting in the receive buffer and return the byte if one is there,
   does NOT block (never waits).

   Parameters: none
   return:     If no byte, then return(-1).  If byte, then return(byte).

   example usage: serial.RxCheck

   expected outcome of example usage call: Return a byte if one is available, but dont wait
                                           for a byte to come in.
}}

  rxByte--                                              'make rxbyte = -1
  if rx_tail <> rx_head                                 'if a byte is in the buffer, then
    rxByte := BYTE[rx_buffer_ptr + rx_tail]             '  grab it and store in rxByte
    'rx_tail := (rx_tail + 1) & (rx_bufsiz - 1)  ' RX_BUFFER_MASK              '  advance the buffer pointer
    ' Because the PASM code uses the rx_tail value to determine if the
    ' receive FiFo is full, we don't want to set it to an invalid value, even
    ' briefly.
    if (rx_tail + 1) => rx_bufsiz
      rx_tail~
    else
      rx_tail += 1

'=========================================================================
PUB RxTime(ms) : rxByte | t
{{
   Wait ms milliseconds for a byte to be received 

   Parameters: ms = number of milliseconds to wait for a byte to be received.
   return:     If no byte, then return(-1).  If byte, then return(byte).

   example usage: serial.RxTime(500)

   expected outcome of example usage call: Wait half a second (500 ms) for a byte to be received.
}}

  t := cnt                                              'take note of the current time
  repeat until ((rxByte := RxCheck) => 0) or ((cnt - t) / (clkfreq / 1000) > ms)
  
'=========================================================================
PUB Rx : rxByte
{{
   Receive byte (may wait for byte)
   returns $00..$FF

   Parameters: none
   return:     received byte

   example usage: serial.Rx

   expected outcome of example usage call: Wait until a byte has been received, then return that byte.
}}

  repeat while (rxByte := RxCheck) < 0                  'return the byte, wait while the buffer is empty

'=========================================================================
PUB Tx(txByte) | tx_next
{{
   Places a byte into the transmit buffer for transmission (may wait for room in buffer).

   Parameters: txByte = the byte to be transmitted
   return:     none

   example usage: serial.Tx($0D)

   expected outcome of example usage call: Transmits the byte $0D serially on the txPin
}}

  'tx_next := (tx_head + 1) & (tx_bufsiz - 1)  ' TX_BUFFER_MASK
  tx_next := tx_head + 1
  if tx_next => tx_bufsiz
    tx_next~

  repeat until (tx_tail <> tx_next)                     'wait until the buffer has room                        
    ' Stall
  BYTE[tx_buffer_ptr + tx_head] := txByte
  tx_head := tx_next                                    'advance the buffer's pointer

  if rxtx_mode & %1000                                  'if ignoring rx echo
    Rx                                                  '   receive the echoed byte and discard

'=========================================================================
PUB Str(stringPtr)
{{
   Transmit a string of bytes

   Parameters: stringPtr = the pointer address of the null-terminated string to be sent
   return:     none

   example usage: serial.Str(@test_string)

   expected outcome of example usage call: Transmits each byte of a string at the address some_string.
}}

  repeat strsize(stringPtr)
    Tx(byte[stringPtr++])                               'Transmit each byte in the string
    
'=========================================================================
PUB Dec(value) | i, x
{{
   Transmit the ASCII string equivalent of a decimal value

   Parameters: dec = the numeric value to be transmitted
   return:     none

   example usage: serial.Dec(-1_234_567_890)

   expected outcome of example usage call: Will print the string "-1234567890" to a listening terminal.
}}

  x := value == NEGX                                    'Check for max negative
  if value < 0
    value := ||(value+x)                                'If negative, make positive; adjust for max negative
    Tx("-")                                             'and output sign

  i := 1_000_000_000                                    'Initialize divisor

  repeat 10                                             'Loop for 10 digits
    if value => i                                                               
      Tx(value / i + "0" + x*(i == 1))                  'If non-zero digit, output digit; adjust for max negative
      value //= i                                       'and digit from value
      result~~                                          'flag non-zero found
    elseif result or i == 1
      Tx("0")                                           'If zero digit (or only digit) output it
    i /= 10                                             'Update divisor

'=========================================================================
PUB Hex(value, digits)
{{
   Transmit the ASCII string equivalent of a hexadecimal number

   Parameters: value = the numeric hex value to be transmitted
               digits = the number of hex digits to print                 
   return:     none

   example usage: serial.Hex($AA_FF_43_21, 8)

   expected outcome of example usage call: Will print the string "AAFF4321" to a listening terminal.
}}

  value <<= (8 - digits) << 2
  repeat digits                                         'do it for the number of hex digits being transmitted
    Tx(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))'  Transmit the ASCII value of the hex characters

'=========================================================================
PUB Bin(value, digits)
{{
   Transmit the ASCII string equivalent of a binary number
   
   Parameters: value = the numeric binary value to be transmitted
               digits = the number of binary digits to print                 
   return:     none

   example usage: serial.Bin(%1110_0011_0000_1100_1111_1010_0101_1111, 32)

   expected outcome of example usage call: Will print the string "11100011000011001111101001011111" to a listening terminal.
}}

  value <<= 32 - digits
  repeat digits
    Tx((value <-= 1) & 1 + "0")                         'Transmit the ASCII value of each binary digit


{{
+-------------------------------------------------------------------------------------------------------------------------------+
|                                                   TERMS OF USE: MIT License                                                  |
+-------------------------------------------------------------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    |
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
+-------------------------------------------------------------------------------------------------------------------------------+
}}