{{
Object file:    Db_FullDuplexSerial_pasm.spin
Version:        1.2.1*
Date:           2006 - 2024
Author:         Chip Gracey, Jeff Martin, Daniel Harris, Tom Dinger
Company:        Parallax Semiconductor
Email:          dharris@parallaxsemiconductor.com
                propeller@tomdinger.net
Licensing:      MIT License - see end of file for terms of use.

Description:
The PASM code and Cog loading code for a single serial port.

Revision History:
       -02/23/2024 [TD] Merged the "best" of the _rr004 version with the -wCTS
                   version; other changes:
                   - Separate tx and rx buffer pointers and sizes in the
                     parameter block
                   - Removed the power-of-two requirement for buffer sizes.
                     (the PASM code doesn't care)
                   - removed the Propeller font extended characters, to make
                     the file ASCII-friendly.
                   - This file contains just the PASM code and the init code
                     to load the PASM cog.
                   TODO: This is now different enough, perhaps I should change
                   the version numbering...

       -12/03/2015 Changed "ifnot ctsPinSet" to "if !ctsPinSet" to correct
                   erroneous use of pin 0 as CTS pin when not explicitly set
       - 2/14/2015 Moved ctsPinSet to DAT with initial zero value
       - 2/26/2014 Modified to permit optional single sided flow control
                  (transmit contingent on CTS high)
       - 1/20/2013 Modified to permit different buffer sizes by incorporating
                    code from Parallax Serial Terminal.
v1.2.1 - 5/1/2011 Added extra comments and demonstration code to bring up
                   to gold standard.
v1.2 - 5/7/2009 Fixed bug in dec method causing largest negative value
                (-2,147,483,648) to be output as -0.
v1.1 - 3/1/2006 First official release.

 
=============================================
        Connection Diagram
=============================================

        +---------+
        |         |         
        |    rxPin|---< TTL level RX line
        |    txPin|---> TTL level TX line
        |   ctsPin|---< CTS line (optional)
        |         |         
        +---------+
         Propeller
            MCU
          (P8X32A)

Components:
N/A

=============================================                  
}}
CON
  ' The buffer sizes do not need to be powers of two, but must be > 1.
  ' The Spin code is a bit slower, but the PASM code is not slowed down at all.
  TX_BUFFERSIZE    = 200
  RX_BUFFERSIZE    = 39

  NO_CTS           = $1E0 '[MMO]   ' [TD] Must have a value > 32 to indicate "not used"

'=========================================================================

PUB LoadCog(rxPin, txPin, mode, baudrate, ctspin, p)
{{
   Start serial driver - starts a cog, but may stall during cog init, until
   the rx and tx buffer address is provided.

   Parameters: rxPin    = Propeller pin to set up as RX-ing pin.  Range = 0 - 31
               txPin    = Propeller pin to set up as TX-ing pin.  Range = 0 - 31
               mode     = bitwise mode configuration variable, see mode bit description below.
               baudrate = baud rate to transmit bits at.
               ctspin   = the pin for CTS, or the value NO_CTS
               p        = address of the 12-long parameter block in the main driver
   
   mode bit 0 = invert rx
   mode bit 1 = invert tx
   mode bit 2 = open-drain/source tx
   mode bit 3 = ignore tx echo on rx

   return: Numeric value of the cog(1-8) that was started, false(0) if no cog is available.
   Note that the Cog will pause, waiting for the rx buffer ptr to be supplied, at
   which point it will continue by collecting the tx buffer ptr and the two buffer
   sizes, and the will start handling the serial line.

   example usage: serial.LoadCog(31, 30, %0000, 9_600)

   expected outcome of example usage call: Starts a serial port on Propller pins 30 and 31.
                                           The serial port does not invert the RX and TX data,
                                           no open-drain/source on the TX pin, does not ignore
                                           data echoed on RX pin, at 9,600 baud.
}}
  ' NOTE: This _must_ match both the layout of variables in Db_FullDuplexSerial
  ' and the presumed layout implemented in the PASM code.

  WORDFILL(p, 0, 13)               ' rx_head, rx_tail, tx_head, tx_tail
                                   ' (and rx_buffer_ptr, tx_buffer_ptr)
  WORD    [p + ( 4 * 2)] := rxPin
  WORD    [p + ( 5 * 2)] := txPin
  WORD    [p + ( 6 * 2)] := mode
  WORD    [p + ( 7 * 2)] := clkfreq / baudrate  ' bit_ticks: clock ticks per bit
  ' rx_buffer_ptr -- filled in later
  ' tx_buffer_ptr -- filled in later
  WORD    [p + (10 * 2)] := RX_BUFFERSIZE
  WORD    [p + (11 * 2)] := TX_BUFFERSIZE
  WORD    [p + (12 * 2)] := ctspin
  return cognew(@entry, p) + 1           'start the new cog now, assembly cog at "entry" label.

'=========================================================================

DAT

'***********************************
'* Assembly language serial driver *
'***********************************

                        org
'
'
' Entry
'
entry                   mov     t1,par                'get structure address
                        add     t1,#4 << 1            'skip past heads and tails

                        rdword  t2,t1                 'get rx_pin
                        mov     rxmask,#1
                        shl     rxmask,t2

                        add     t1,#2                 'get tx_pin
                        rdword  t2,t1
                        mov     txmask,#1
                        shl     txmask,t2

                        add     t1,#2                 'get rxtx_mode
                        rdword  rxtxmode,t1

                        add     t1,#2                 'get bit_ticks
                        rdword  bitticks,t1

                        add     t1,#2                'get rx_buffer_ptr
                        ' Wait for the buffer pointer to be supplied.
:bufwait
                        rdword  rxbuff,t1  wz
        if_z            jmp     #:bufwait

                        add     t1,#2                 ' get tx_buffer_ptr
                        rdword  txbuff,t1

                        add     t1,#2                 ' get the rx bufsiz
                        rdword  rxbufsiz,t1           ' currently the size

                        add     t1,#2                 ' get the tx bufsize
                        rdword  txbufsiz,t1

                      '[MMO]
                        add     t1,#2                 ' get cts_pin address
                        rdword  t2,t1
                        mov     ctsmask,#0            '
                        cmp     t2,#32          wc    ' 
        if_c            mov     ctsmask,#1
        if_c            shl     ctsmask,t2                        
                      '''''
                        test    rxtxmode,#%100  wz    'init tx pin according to mode
                        test    rxtxmode,#%010  wc
        if_z_ne_c       or      outa,txmask
        if_z            or      dira,txmask

                        mov     txcode,#transmit      'initialize ping-pong multitasking
'
'
' Receive
'
receive                 jmpret  rxcode,txcode         'run chunk of tx code, then return

                        test    rxtxmode,#%001  wz    'wait for start bit on rx pin
                        test    rxmask,ina      wc
        if_z_eq_c       jmp     #receive

                        mov     rxbits,#9             'ready to receive byte
                        mov     rxcnt,bitticks
                        shr     rxcnt,#1
                        add     rxcnt,cnt                          

:bit                    add     rxcnt,bitticks        'ready next bit period

:wait                   jmpret  rxcode,txcode         'run chunk of tx code, then return

                        mov     t1,rxcnt              'check if bit receive period done
                        sub     t1,cnt
                        cmps    t1,#0           wc
        if_nc           jmp     #:wait

                        test    rxmask,ina      wc    'receive bit on rx pin
                        rcr     rxdata,#1
                        djnz    rxbits,#:bit

                        shr     rxdata,#32-9          'justify and trim received byte
                        and     rxdata,#$FF
                        test    rxtxmode,#%001  wz    'if rx inverted, invert byte
        if_nz           xor     rxdata,#$FF

                        rdword  t2,par                'save received byte and inc head
                        add     t2,rxbuff
                        wrbyte  rxdata,t2
                        sub     t2,rxbuff
                        add     t2,#1
                        cmpsub  t2,rxbufsiz
                        wrword  t2,par

                        jmp     #receive              'byte done, receive next byte
'
'
' Transmit
'
transmit                jmpret  txcode,rxcode         'run chunk of rx code, then return
        
                        ''''' 
                        '[MMO]
        ' if CTS is in use, await CTS.. tx function will pause if buffer is filled while waiting
                        cmp     ctsmask,#0      wz
        if_z            jmp     #:ignoreCTS
                        test    ctsmask,ina     wz    ' Z set if CTS is low
        if_z            jmp     #transmit             ' we are not clear to send
:ignoreCTS 
        ' [TD] TODO: We can get rid of one instruction here...
        ' Well, if we are using CTS it is one instruction shorter, but if we are
        ' not using cts, then it is one instruction _longer_. The only way to fix
        ' that would probably be some kind of code rewriting at runtime in the
        ' initialization code.
        '                cmp     ctsmask,#0      wz    ' Z clear if we honor CTS
        '                test    ctsmask,INA     wc    ' C clear if CTS is not asserted
        'if_nz_and_nc    jmp     #transmit             ' CTS is not ready, and we pay
                                                      ' attention to it.
                        '''''
                        mov     t1,par                'check for head <> tail
                        add     t1,#2 << 1
                        rdword  t2,t1
                        add     t1,#1 << 1
                        rdword  t3,t1
                        cmp     t2,t3           wz
        if_z            jmp     #transmit             ' queue empty.. return to top

                        add     t3,txbuff             'get byte and inc tail
                        rdbyte  txdata,t3
                        sub     t3,txbuff
                        add     t3,#1
                        cmpsub  t3,txbufsiz
                        wrword  t3,t1

                        or      txdata,#$100          'ready byte to transmit
                        shl     txdata,#2
                        or      txdata,#1
                        mov     txbits,#11
                        mov     txcnt,cnt

:bit                    test    rxtxmode,#%100  wz    'output bit on tx pin 
                        test    rxtxmode,#%010  wc    'according to mode
        if_z_and_c      xor     txdata,#1
                        shr     txdata,#1       wc
        if_z            muxc    outa,txmask        
        if_nz           muxnc   dira,txmask
                        add     txcnt,bitticks        'ready next cnt

:wait                   jmpret  txcode,rxcode         'run chunk of rx code, then return

                        mov     t1,txcnt              'check if bit transmit period done
                        sub     t1,cnt
                        cmps    t1,#0           wc
        if_nc           jmp     #:wait

                        djnz    txbits,#:bit          'another bit to transmit?

                        jmp     #transmit             'byte done, transmit next byte
'
'
'
ctsPinSet               byte    false                 ' MMO
'
' Uninitialized data
'
t1                      res     1
t2                      res     1
t3                      res     1

rxtxmode                res     1
bitticks                res     1

rxmask                  res     1
rxbuff                  res     1
rxdata                  res     1
rxbits                  res     1
rxcnt                   res     1
rxcode                  res     1
rxbufsiz                res     1

txmask                  res     1
txbuff                  res     1
txdata                  res     1
txbits                  res     1
txcnt                   res     1
txcode                  res     1
txbufsiz                res     1

ctsmask                 res     1               ' MMO

                        FIT 496                 ' added for consistency
                        
{{
+-------------------------------------------------------------------------------------------------------------------------------+
|                                                   TERMS OF USE: MIT License                                                  |
+-------------------------------------------------------------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    |
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
+-------------------------------------------------------------------------------------------------------------------------------+
}}