{{
HUB75 panel driver configuration object

Author:         Tom Dinger (Ding-Batty)
Release Date:   2024/02/21

Copyright 2024 Thomas E. Dinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

--------------------------------------------------------------------------

This file contains the configuration constants for the Hub75-based display
hardware as it is connected to the Propeller. It contains pin assignments,
display geometry, color depth (number of bits per color value), etc.

This file also contains some other helpful constants that are derived from
the configuration settings (a form of common subexpression elimination).

The current pin assignments are for my use of the Propeller Board of Education,
along with one of the "P2 Eval HUB75 Adapter Boards" designed by Stephen Moraco
and sold by Parallax. 

For more information:
---------------------

See the file "hub75_info.txt" for some details on the display interface
called "HUB75", with some useful links.

}}

CON
  {
    Display Geometry
  }
  ' DISPLAY_ROWS is the number of rows physically on the display. This is a
  ' power of two, usually 16, 32, or 64.
  '
  ' For HUB75 type displays, rows are paired for output on the display, so two
  ' rows are written at a time, so the number of refresh "rows" is
  ' DISPLAY_ROWS / 2.

  DISPLAY_ROWS = 64  ' Panels are all 64 rows

  ' DISPLAY_COLUMNS is the number of LEDs per row in the daisy-chained set
  ' of displays, so it might not be a power of two. However, note that if
  ' this is set to a number smaller than the actual number of LED columns,
  ' the columns "left out" on the left of the display will get "junk"
  ' displayed from the right edge of the previous row output (ask me how I
  ' know :).

  DISPLAY_COLUMNS = 64 * 2 ' Two 64 x 64 panels in series

  {
   Pin assignments
  }
  ' COLORS_BASE_PIN is the lowest pin number of the six color pins used for
  ' output.
  '
  ' The color order within the three color bits for a pixel don't matter, but
  ' depend on the wiring to the display panel, and need to be specified
  ' below for correct data conversions. However, the two groups of three
  ' bits should have the same color order, e.g. RGBrgb, or GRBgrb, etc.
  '
  ' For the hub75_display_pasm_waitcnt.spin and hub75_display_pasm_repeat.spin
  ' drivers, the COLORS_BASE PIN should be 0 or 9 corresponding to the target
  ' of the MOVS and MOVD instructions, respectively. (Base pin 23 is not really
  ' usable, because it runs into the I2C pins.)
  '
  ' For the hub75_display_pasm_video_8clock.spin driver, the base pin is the
  ' base pin of the video groups: 0, 8, or 16 (24 is not available for the
  ' same reason as pin 23 for MOVI, above).
  COLORS_BASE_PIN  = 0

  ' ROW_NUMBER_BASE_PIN is the lowest pin number for the row selection bits.
  ' The number of row selection bits is derived from the value of DISPLAY_ROWS
  ' set above.
  ROW_NUMBER_BASE_PIN = 9

  ' The other control bits

  SHIFT_CLOCK_PIN = 15

  ' We unify the ShiftLatch and OEInverted pins -- they each latch on the
  ' opposite edges:
  ' - ShiftLatch latches on the rising edge
  ' - OEInverted latches on the falling edge (the enabling edge).
  ' So, in theory, the ShiftLatch can be used to latch data in the first latch
  ' before the output is displayed with the second "OE latch". (At least, for
  ' for the boards using the ICN2037, which is what I have.)
  ' As implemented, the code will work either with two separate pins, or
  ' with both functions sharing a pin.
  SHIFT_LATCH_PIN    = 14
  OE_INVERTED_PIN    = 14
  SHIFT_LATCH_OE_PIN = 14

  ' Color bit ordering.
  ' In order to be able to convert separate r, g, b multibit values into
  ' the groups of three RGB bit based on bit priority, and to match the
  ' expected pin ordering for the six color output pins.

  RED_BIT      = %100
  GREEN_BIT    = %010
  BLUE_BIT     = %001

  ' We have two formats for a color scalar:
  ' - Simple RGB values, one color per byte, as in 0RGB
  ' - Reordered color bits for fast display updates, e.g.
  '   fxxxxxxxrgbRGBrgbRGBrgbRGBrgbRGBrgbRGBrgbRGB
  ' The high-order bit of the color determines which format the color value
  ' uses:
  ' - If the high bit is clear, it is imply 0RGB
  ' - If the high bit is set, it has been converted to display color bit order.
  DISPLAY_COLOR_BIT_ORDER = $8000_0000

  ' The display refresh routine has a preferred order for sending color bits
  ' to the display, for managing intensity levels. Some want to send the low-
  ' order bit first, some want to send the high order bit first.
  ' We want to represent enough information here for the code that writes into
  ' the display buffer to be able to adapt to the preferred order, with a small
  ' or no real affect on efficiency, and without needing multiple versions of
  ' the buffer write code in any given binary.
  '
  ' So we provide the starting offset of the low-order bit color layer in the
  ' display buffer, and the offset of the high order bit layer in the buffer,
  ' and provide the constants used to walk from one to the other. Then the
  ' code can walk in either preferred direction to write the color bits.
  ' - display low-order color layer offset
  ' - display high-order color layer offset
  ' - layer address delta going low to high
  ' - layer address delta going high to low
  '
  ' This will allow a single "Display color format" representation to be used
  ' for display refresh routines that prefer either order, with no code changes,
  ' just constant changes that control the address calculations.
  '
  ' Since we don't want to require specifying the COLOR_DEPTH value here (it
  ' is application-specific, not hardware specific), we can only specify the
  ' delta walking from low ordee layer to higher, and vice versa, and then
  ' use that at runtime to determine the display buffer addresses for the lowest
  ' and highest color layers.

  ' Display refresh that sends low-order color layers first:
  ' PropTool 1.3.2 doesn't like names longer than 30 characters
  'DISPLAY_BUFFER_LOW_TO_HIGH_LAYER_DELTA =   DISPLAY_COLUMNS * REFRESH_ROWS
  'DISPLAY_BUFFER_HIGH_TO_LOW_LAYER_DELTA = -(DISPLAY_COLUMNS * REFRESH_ROWS)
  DBUF_LO2HI_LAYER_DELTA =   DISPLAY_COLUMNS * REFRESH_ROWS
  DBUF_HI2LO_LAYER_DELTA = -(DISPLAY_COLUMNS * REFRESH_ROWS)
  
  '' Display refresh that send high-order color layers first: 
  ''DISPLAY_BUFFER_LOW_TO_HIGH_LAYER_DELTA = -(DISPLAY_COLUMNS * REFRESH_ROWS)
  ''DISPLAY_BUFFER_HIGH_TO_LOW_LAYER_DELTA =   DISPLAY_COLUMNS * REFRESH_ROWS
  'DBUF_LO2HI_LAYER_DELTA = -(DISPLAY_COLUMNS * REFRESH_ROWS)
  'DBUF_HI2LO_LAYER_DELTA =   DISPLAY_COLUMNS * REFRESH_ROWS

' We will use a "FIT" operation trick in the PASM code to validate that
' the configuration is valid, such as making sure the pin assignments are
' sensible, the sizes are within limits, etc.
'
' If we can be sure that a DAT section here with no contents takes up no
' actual HUB memory, then we can do it here, instead of in the PASM file.

' ========================================================================
{ Derived constants
  These are useful constants for various purposes derived from the base
  configuration.
}
  MAX_DISPLAY_ROW      = DISPLAY_ROWS - 1
  REFRESH_ROWS         = DISPLAY_ROWS >> 1
  MAX_REFRESH_ROW      = REFRESH_ROWS - 1
  ROW_NUMBER_BITS      = (>| REFRESH_ROWS) - 1 ' 5 for a 32-row refresh
  ROW_NUMBER_MASK_BASE = MAX_REFRESH_ROW

  MAX_DISPLAY_COLUMN   = DISPLAY_COLUMNS - 1

  COLORS_SHIFT         = COLORS_BASE_PIN
  COLORS_MASK          = %111_111 << COLORS_SHIFT

  ROW_NUMBER_MASK      = ROW_NUMBER_MASK_BASE << ROW_NUMBER_BASE_PIN
  ROW_NUMBER_INCREMENT = |< ROW_NUMBER_BASE_PIN

  SHIFT_CLOCK_MASK     = |< SHIFT_CLOCK_PIN
  ' All the same
  SHIFT_LATCH_MASK     = |< SHIFT_LATCH_PIN
  OE_INVERTED_MASK     = |< OE_INVERTED_PIN
  SHIFT_LATCH_OE_MASK  = SHIFT_LATCH_MASK | OE_INVERTED_MASK

  ' These may not be needed.
  RED_BIT_SHIFT        = (>| RED_BIT) - 1
  GREEN_BIT_SHIFT      = (>| GREEN_BIT) - 1
  BLUE_BIT_SHIFT       = (>| BLUE_BIT) - 1

PUB NoOp
  ' Needed for a "proper" object
