{{
HUB75 panel driver obect, with a PASM display refresh cog.

Author:         Tom Dinger (Ding-Batty)
Release Date:   2024/02/22

Copyright 2024 Thomas E. Dinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

--------------------------------------------------------------------------

This file contains the access routines for one or more HUB75 display panel(s).
It uses PASM code to continuously refresh the display from a display buffer
stored in HUB memory. It contains write routines to update the display buffer
contents. (The pasm code is in a separate file.)

This is mostly statically configured, using settings from the hub75_config.spin
file; the final initialization is deferred until after the cog is started,
so that the hub memory that contained the PASM code loaded into the cog can
be recovered and used for the display buffer, and other buffers.

This file contains a lot of functions in Spin to write to the display buffer,
but if you are going to use the PASM-based write helper object, then you don't
really need all the Spin-based write routines.

For more information:
---------------------

See the file "hub75_info.txt" for some details on the display interface
called "HUB75", with some useful links.

Notes about PASM Cog start-up
-----------------------------

A separate cog running PASM code does the actual work of refreshing the display
from the in-memory display buffer. It has a two-step start-up:
1. Most of the runtime configuration values, if not constants, are "poked"
   into the hub copy of the PASM code for the COG; the exception is the base
   address of the display buffer in hub memory.

2. The cog is started.

3. The code in the Cog waits for the hub variable pointed to by the PAR value
   at cog start, to contain a non-zero value, which is taken to be the base
   address of the display buffer in hub memory. So, the COG will stall very
   early in start-up until that address is provided.

   This delay allows the controlling program to reclaim the hub memory copy of
   the cog code, for re-use as part of the (large) display buffer.

4. Once all the cogs have been started, and the memory for the various buffers
   has been laid out (and possibly initialized), the address of the display
   buffer is written to the location being queried by the cog, allowing it to
   complete initialization and start refreshing the display.

The helper routines that also use a PASM cog to speed up writing into the
display buffer use a similar two-step start-up process.

With proper control over memory layout of the Spin object, by putting all the
PASM code in one area, that entire area can be reclaimed for use as buffers.

Display Buffer memory layout
============================

This driver assumes the following memory organization:

Each BYTE of the display buffer contains 3 color bits for each of two RGB LEDs
in the same column, but REFRESH_ROWS apart. The six color bits are stored in
the low-order bits of the byte: %xx_RGB_rgb (for whichever color bit order is
configured). These are written to OUTA using the video generator in the cog
to write out the six bis using the VGA output mode.

The color values for one row are stored in consecutive bytes, with the lower-
address byte containing the lower columns. Rows are also stored consecutively,
with the rows at the top of the display at the lower addresses, so the LED
for row == 0 and column == 0 is the upper-left corner LED.

A Columns x Rows display will use (Columns + Rows) / 2 bytes of storage.
So, a 32 x 32 display will use 512 bytes of storage, and a 64 x 64 display
will use 2048 bytes of storage, for one color bit.

For color depth > 1 (more than 1 bit per color), the bits are stored in
"layers" with one layer for each color bit. These layers are also stored
consecutively, with the lowest-order color bit first (this can change.)

The memory layout for the color bits makes it easier to refresh the display,
but likely requires help with writing into the display buffer from a write-
helper cog, as Spin is too slow for writes of any large numbers of pixels.


Performance
===========

To be refined.

In Spin only, with 80MHz system clock, we can get about 17 panel refreshes per
second, with one-bit color depth. Pretty poor -- there is a lot of flickering
on the display.

The first version of the PASM driver, with no significant optimizations (other
than using a MOVD instruction to set the colors, instead of two (or more)
instructions, gets about 1192 refreshes per second (one-bit color depth).
We can more then double that with further optimizations.

With the optimization of reading LONGs instead of BYTEs (4 columns at a time),
and using a simple pulse output from a CTR, we reach 2_328.8 single-bit
refreshes a second for a single panel. (8 bits would be 2329/255 == 9.133
refreshes per seond for a single panel. 4 bits would be 2329/15 == 155
refreshes per second.)

Further, using both counters to generate the burst of four shift pulses, we
reach 2_942.2 refreshes per second for a single panel. (Four bits would be
2942.2/15 = 196 refreshes per second; Six bits: 46.7 refreshes per second;
Eight bits: 11.5 refreshes per second. These numbers will be a bit slower
with the per-color-bit overhead.)

Even further, using the video generator to do the bit-shifting, we can get
a 50% increase in refresh rate at 8 sysclks per pixel, and may even be able
to roughly double that.

Detailed performance numbers: refresh rates under about 60Hz are not really
usable, because there is too much visible flicker. YMMV.
- Refresh rates: 64x64 display
  Color   Waitcnt   Repeat    Vid8clk
  Depth   Rate      Rate      Rate
  -----   ------    ------    -------
    1     2886.8    2942.5    4340.3
    2      963.8     988.1    1446.8
    3      413.2     425.6     620.0
    4      192.9     199.2     289.4
    5       93.3      96.6     140.0
    6       45.9      47.6      68.9
    7       22.8      23.6      34.2
    8       11.3      11.8      17.0
- Refresh rates: 128x64 display
  COLOR_DEPTH=8 is not available because the display buffer would take up _all_
  of hub ram: 32KiB
  Color   Waitcnt   Repeat    Vid8clk
  Depth   Rate      Rate      Rate
  -----   ------    ------    -------
    1     1530.3    1545.5    2297.8
    2      510.5     517.2     765.9
    3      218.8     222.2     328.3
    4      102.1     103.9     153.2
    5       49.4      50.3      74.1
    6       24.3      24.8      36.5
    7       12.1      12.3      18.1
    8       n/a       n/a       n/a

Different display configurations vary linearly with the number of display
columns; For refresh rows of 8 or 16 (the above is 32), the refresh rate
will roughly double or quadruple. This could be measured using the test
program measure_refresh_rate.spin, and varying the config in hub75_config.spin.
(No actual display is required for the refresh measurements.)
}}

OBJ
  ' Pick up the static configuration.
  Config: "hub75_config.spin"

CON
  ' Import the needed configuration settings from the Config object, for
  ' code brevity.
  REFRESH_ROWS         = Config#REFRESH_ROWS
  MAX_REFRESH_ROW      = Config#MAX_REFRESH_ROW
  DISPLAY_ROWS         = Config#DISPLAY_ROWS
  MAX_DISPLAY_ROW      = Config#MAX_DISPLAY_ROW
  DISPLAY_COLUMNS      = Config#DISPLAY_COLUMNS
  MAX_DISPLAY_COLUMN   = Config#MAX_DISPLAY_COLUMN
  SHIFT_CLOCK_PIN      = Config#SHIFT_CLOCK_PIN
  SHIFT_LATCH_PIN      = Config#SHIFT_LATCH_PIN
  OE_INVERTED_PIN      = Config#OE_INVERTED_PIN
  SHIFT_CLOCK_MASK     = Config#SHIFT_CLOCK_MASK
  ' Note: ShiftLatch and OEInverted are the same pin in this version, but they
  ' do not need to be the same.
  SHIFT_LATCH_MASK     = Config#SHIFT_LATCH_MASK
  OE_INVERTED_MASK     = Config#OE_INVERTED_MASK
  ROW_NUMBER_MASK      = Config#ROW_NUMBER_MASK
  ROW_NUMBER_INCREMENT = Config#ROW_NUMBER_INCREMENT
  DBUF_LO2HI_LAYER_DELTA = Config#DBUF_LO2HI_LAYER_DELTA
  DBUF_HI2LO_LAYER_DELTA = Config#DBUF_HI2LO_LAYER_DELTA

  BYTES_PER_COLOR_LAYER = REFRESH_ROWS * DISPLAY_COLUMNS

  RED_BIT      = Config#RED_BIT
  GREEN_BIT    = Config#GREEN_BIT
  BLUE_BIT     = Config#BLUE_BIT

  RED_BIT_SHIFT   = Config#RED_BIT_SHIFT
  GREEN_BIT_SHIFT = Config#GREEN_BIT_SHIFT
  BLUE_BIT_SHIFT  = Config#BLUE_BIT_SHIFT

VAR
  LONG    DisplayCog      ' 1 + Cog# for display refresh cog
  LONG    ColorDepthBits  ' Number of color layers > 0
  LONG    MaxColorValue   ' set to (|< ColorDepthBits) - 1

  ' The display buffer base address in hub memory
  LONG    BufAddrHub 

  ' These point to the lowest and highest layer for color bit output, so
  ' piixel writes can work for either layer order, along with the
  ' delta constants to move between layers in the display buffer.
  LONG    DisplayBufferHighLayerAddr
  LONG    DisplayBufferLowLayerAddr

  ' FullRefreshClocks gets the count of sysclks needed for one full refresh of
  ' all the color bits for the display, so that we can report on the effective
  ' refresh rate, depending on the size of the display and the color depth. 
  LONG    FullRefreshClocks

' ========================================================================
PUB Stop
{{
  Stops the PASM cog, if it is running.
  Also disables the display output.

  Arguments:
    None

  Return Value:
    None
}}
  if DisplayCog
    ' Turn off the display
    OUTA[OE_INVERTED_PIN]~~   ' High means disabled
    DIRA[OE_INVERTED_PIN]~~   ' Now an output, so display is disabled.

    ' Strictly speaking, there is a pull-up on the OEInverted pin in the
    ' display itself, so simply letting the pin float as an input should be
    ' sufficient to diable output. But using a level shifter will always
    ' drive the 5V side, so we have to make sure that the Propeller side of
    ' the level shifter has a pull-up as well. If it does, all we really
    ' to do is to set the OEInvertedPin high (as an output pin), and once
    ' the cog is fully stopped (should be immediate on the return of the
    ' COGSTOP() operation) we can then float the OEInvertedPin again.

    ' This overrides what the display cog is doing, forcing the display
    ' output to be disabled. So we are free to stop the cog.
    ' (Ideally, we'd also shift out a full row of zeros, so that even if
    ' the OE bit flips during reset, the data displayed is all zeros, so
    ' there will be no flash on the display.)
    COGSTOP(DisplayCog~ - 1)

    ' Consider floating the OEInvertedPin again here. (Perhaps test for
    ' shutdown flashes?)

' ========================================================================
PUB GetDisplayBufAddrVar
{{
  Gets the address in hub memory of the communication variable that the PASM
  cog code examines, waiting for it to become non-zero -- the non-zero value
  is the base address of the display buffer in hub.

  Arguments:
    None

  Return Value:
    None
}}
  return @BufAddrHub

' ========================================================================
PUB GetRefreshClocksAddrHub
{{
  Gets the address in hub memory of the communication variable that the PASM
  cog code puts the sysclk count for an entire refresh pass over all color
  layers of the display. This count is used to calculate the display refresh
  rate specific for the configured display dimensions and color depth.

  Arguments:
    None

  Return Value:
    None
}}
  return @FullRefreshClocks

' ========================================================================
PUB FinalDisplayInit(CogPlusOne, BufAdr, ColorDepth)
{{
  Finishes the initialization for display refresh:
  - records the running cog number of the display refresh code
  - records final configuration information (display buffer address, color
    depth)
  - notifies the display refresh cog of the buffer address so that it can
    conplete its own init and start refreshing the display.

  Arguments:
    CogPlusOne - the cog ID of the display refresh cog, plus one, so that if
                 the saved value is zero, the cog is not (yet) running.
    BufAdr - the address in hub memory of the display buffer.
    ColorDepth - the number of bits per color for the display. The display
                 buffer stores the color information in "layers", one layer
                 for each bit of color depth.

  Return Value:
    None
}}
  DisplayCog     := CogPlusOne
  ColorDepthBits := ColorDepth
  MaxColorValue  := (|< ColorDepth) - 1

  ' We also calculate the low-order color bit layer, and the high order one
  ' so we can propery handle mapping a display-color-order color to the
  ' layers.

  ' Sigh -- BST has some problems with some optimizations levels for constant
  ' expressions involving the numeric comparisions < =< => >, and can be seen
  ' even with optimizations turned off, if using the constant() form.
  ' It doesn't seem to see contants as negative values, but I have not fully
  ' characterized the problem.
  ' So, this negative-number test is a work-around for that problem.
  'if DBUF_LO2HI_LAYER_DELTA < 0
  if (DBUF_LO2HI_LAYER_DELTA & $80000000) <> 0
    ' The display buffer starts with the high-order color bit layer
    DisplayBufferHighLayerAddr := BufAdr
    DisplayBufferLowLayerAddr := BufAdr + (BYTES_PER_COLOR_LAYER * (ColorDepth - 1))
  else
    ' The display buffer starts with the low-order color bit layer
    DisplayBufferLowLayerAddr := BufAdr
    DisplayBufferHighLayerAddr := BufAdr + (BYTES_PER_COLOR_LAYER * (ColorDepth - 1))

  ' The Display refresh cog should already be running, so all we need to do
  ' is to tell it the address of the display buffer in hub memory.
  BufAddrHub     := BufAdr

' ========================================================================
PUB GetRefreshClocks | n
{{
  Returns the number of sysclks taken for one full refresh of the display.
  This will stall until the number (non-zero) is available from the PASM cog.

  Arguments:
    None.

  Return Value:
    The sysclk count for one full display refresh. This can be used to
    calculate the display refresh rate.
}}
  repeat while FullRefreshClocks == 0
    ' Stall until we know we have the count
  return FullRefreshClocks

' ========================================================================
{{
  Color conversions.

  We support up to 8 bits per color for the display, depending on the display
  size (More color bits and larger sizes may result in refresh flickering.)
  (We could support up to 10 bits of color, with some changes.)

  The "normal" RGB storage style, e.g. one color per byte, is somewhat
  inconvenient for writing to the display buffer because of the color layers,
  one layer per color bit.

  So we have two formats for a color scalar:
  - Simple RGB values, one color per byte, as in $00RRGGBB
  - Reordered color bits for fast display updates, e.g.
    %fxxxxxxxrgbRGBrgbRGBrgbRGBrgbRGB
    Where the low order bits are the low order color bits.
  The high-order bit of the color determines which format the color value
  uses:
  - If the high bit is clear, it is imply $00RRGGBB
  - If the high bit is set, it has been converted to display color bit order.

  We have convenience functions to convert to display color format, which for
  frequently used colors in an application saves time when writing to the
  display buffer by storing the converted color and using that.
}}
' ---------------------------------------
PUB RgbColor(r, g, b)
{{
  Converts the three color values into a single LONG color value, with
  one color per byte, in the format $00RRGGBB.

  No range-checking is done to limit the color values to the color depth
  cofigured for the display, other than limiting the values to 0..255 by
  masking.

  Arguments:
    r -- the red value, 0..255
    g -- the greeen value, 0..255
    b -- the blue value, 0..255

  Return Value:
    The combined colors in one long: $00RRGGBB
}}
  return ((r & $FF) << 16) | ((g & $FF) << 8) | (b & $FF)

' ---------------------------------------
PUB DisplayColor(r, g, b)
{{
  Converts the three color values into a single LONG color value in the form
  most conveniently used for writing into the display buffer.

  No range-checking is done to limit the color values to the color depth
  cofigured for the display, other than limiting the values to 0..255 by
  masking.

  Arguments:
    r -- the red value, 0..255
    g -- the greeen value, 0..255
    b -- the blue value, 0..255

  Return Value:
    The combined colors in one long, with a flag indicating it is in display
    color format.

  Someday, this may use 10-bit colors instead of 8 bit.
}}
  return ToDisplayColor(RgbColor(r, g, b))

' ---------------------------------------
PUB ToDisplayColor(color) | c
{{
  Interleave the three color values from a $00RRGGBB format color into sets of
  three bits, and order them for more convenient use for display buffer writes.

  For example, using 3 bits per color (9-bit color), we start with
    r = r2_r1_r0   (bits)
    g = g2_g1_g0
    b = b2_b1_b0
  and the result is
    color = r2_g2_b2_r1_g1_b1_r0_g0_b0
  These are used, three bits at a time, to update the color bit layers
  in the display buffer.

  Display color bit order has the high-order color bits in the low-order bits
  of the color value;' the write routines write the high-order color bits
  first.

  If the color argument already is in the display color format, it is
  simply returned unchanged.

  No range-checking is done to limit the color values to the color depth
  cofigured for the display, other than limiting the values to 0..255 by
  masking.

  Arguments:
    color -- the RGB formatted color value

  Return Value:
    The combined colors in one long, with a flag indicating it is in display
    color format.

  TODO: Consider changing to 10-bits-per-color for dynamic range, even though
  the larger displays cannot be refreshed at larger color depths. The above is
  specifially limited to 8 bits per color. More than 10 bits would require non-
  trivial changes, since the color value for a pixel would not fit into a
  single LONG.
}}
  if (color & Config#DISPLAY_COLOR_BIT_ORDER) <> 0
    return color   ' No changed needed; already in Display color format
  c := 0
  repeat ColorDepthBits
    c <-= 3
    if (color & 1) <> 0
      c |= BLUE_BIT
    color ->= 8
    if (color & 1) <> 0
      c |= GREEN_BIT
    color ->= 8
    if (color & 1) <> 0
      c |= RED_BIT
    color ->= 17
  
  return c | Config#DISPLAY_COLOR_BIT_ORDER

' ---------------------------------------
' TODO: Implement ToRgbColor(c) -- not needed (yet)

' ========================================================================
{{
  Write routines.
  The function signatures here are consistent with those used by the
  PASM-assisted write helper object (hub75_helper.spin).

  If an application will be using the helper object that uses another cog
  for speed, these may not be needed in that application and could be
  removed.
}}

' ---------------------------------------
PUB Clear
{{
  Set all the pixels in the display buffer to black.

  Arguments:
    None.

  Return Value:
    None.
}}
  ' Simply zero-fill the entire hub area
  LONGFILL(BufAddrHub, 0, (BYTES_PER_COLOR_LAYER * ColorDepthBits) >> 2)

' ---------------------------------------
PUB Pixel(col, row, color) | mask, p, colorsh
{{
  Set a single pixel to a particular color

  Arguments:
    col -- the column number, zero-based. The leftmost column is column 0.
    row -- the row number, zero-based, The topmost row is row 0.
    color -- the color to set the pixel. It may either be the RGB color or
             the display color

  Return Value:
    None.
}}
  ' We do byte addressing to do the update.
  color := ToDisplayColor(color)
  mask := %11_111_000   ' RGB, in some order
  colorsh := 0
  if row => REFRESH_ROWS  ' If in the bottom half of the display...
    mask := %11_000_111   ' ... Use bits %111000 instead of bits %000111
    colorsh := 3
    row -= REFRESH_ROWS   ' bottom half is stored overlaying the top half.

  ' We write high-order color bits first
  p := DisplayBufferHighLayerAddr + (row * DISPLAY_COLUMNS) + col
  repeat ColorDepthBits
    BYTE[p] := (BYTE[p] & mask) | ((color & %111) << colorsh)
    color >>= 3  ' move to the RGB bits for the next layer
    p += DBUF_HI2LO_LAYER_DELTA  ' next layer of color bits.

' ---------------------------------------
PRI HorizontalHelper(p, len, bits, shift, mask, fgcolor, bgcolor) | c
{{
  Helper function to write either the foreground color bits or the
  background color bits in a single row of a single display color layer,
  based on a bit pattern to select foreground or background.

  This is used as a helper for drawing horizontal lines, or lines from
  a 1bpp or 2bpp bitmap. 

  Arguments:
    p -- the pointer into the display buffer for the starting point (left end
         of the horizontal pixels to set)
    len -- The number of pixels to set (length of the range of pixels drawn)
    bits -- the bits controlling foreground or background color writing;
            1 means the foreground color. The low-order bits in the value
            are the leftmost pixels in the row.
    shift -- the distance to shift the "bits" value to move to the next
             pixel's bit. It has the value 1 for 1 bit-per-pixel (1bpp),
             or 2 for 2bpp, as is needed for the ROM Font.
    mask -- the color mask to use when updating the bytes in the display 
            buffer; this will be %000111 for a row in the top half of the
            display, and %111000 for the bottom half of the display.
    fgcolor -- the foreground color, either RGB or display formatted.
    bgcolor -- the background color, either RGB or display formatted. 

  Return Value:
    None.
}}
  ' This can also be used to draw solid-color horizontal lines by setting
  ' bits to 0, which will write only using the bgcolor -- the code uses a
  ' right-shift to check each bit, which means that any length will always
  ' use the bgcolor color.
  '
  ' The code assumes that all boundary-checking as already been done, so
  ' that every pixel will be on the display.
  '
  repeat len
    if bits & 1 <> 0
      c := fgcolor
    else
      c := bgcolor
    BYTE[p] := (BYTE[p] & mask) | c
    bits ~>= shift  ' get the next bit
    ++p

PUB HorizontalLine(column, row, len, color) | mask, p, colorsh  ', c
{{
  Draw a horizontal line.

  Arguments:
    column -- the column number for the left end of the horizontal line,
              zero-based. The leftmost column is column 0. It is range-limited
              via masking to 0..MAX_DISPLAY_COLUMN (it probably should be
              via min/max operators instead.)
    row -- the row number, zero-based, The topmost row is row 0. It is
           range-limited via masking to 0..MAX_DISPLAY_ROW
    len -- the length of the line in pixels.
    color -- the color to set the pixel. It may either be the RGB color or
             the display color

  Return Value:
    None.
}}
  row    &= MAX_DISPLAY_ROW
  column &= MAX_DISPLAY_COLUMN
  ' length of run is limited to the right edge of the display
  len <#= DISPLAY_COLUMNS - column

  mask := %11_111_000
  colorsh := 0
  if row => REFRESH_ROWS   ' If in the bottom half of the display...
    mask := %11_000_111    ' ... Use bits %111000 instead of bits %000111
    colorsh := 3
    row -= REFRESH_ROWS
  color := ToDisplayColor(color)

  ' We write high-order color bits first
  p := DisplayBufferHighLayerAddr + (row * DISPLAY_COLUMNS) + column
  repeat ColorDepthBits
    'c := (color & %111) << colorsh
    HorizontalHelper(p, len, 0, 1, mask, 0, (color & %111) << colorsh)
    color >>= 3  ' Drop the used color bits.
    p += DBUF_HI2LO_LAYER_DELTA

' ---------------------------------------
PRI VerticalLineHelper(p, l, color, mask)
{{
  Helper function to write the color bits for one color layer for a line of
  pixels vertically. All the rows should be either in the top half or the
  bottom half of the DISPALY_ROWS, because the color and mask values are
  specifc to the bits in the display buffer bytes used for the two halves
  of the display.

  This is used as a helper for drawing vertical lines.

  Arguments:
    p -- the pointer into the display buffer for the starting point (top end
         of the vertical pixels to set)
    len -- The number of pixels to set (length of the range of pixels drawn).
           The pixels should all be in the same half of the display.
    color -- the three color bits for one layer, in the correct position for
             the half of the display being updated: either %00_000_xxx for the
             top half, or %00_xxx_000 for the bottom half.
    mask -- the mask used to clear the "old" bits from the byte before adding
            in the "color" bits for each pixel: either %11_111_000 for the top
            half, or %11_000_111 for the bottom half.

  Return Value:
    None.
}}
  ' Helper for setting pixels vertically
  ' Assumes that the color and mask values are in the right position for
  ' the set of rows, which should be entirely within the first 32 rows,
  ' or the last 32 rows, but not cross the boundary between them.
  repeat while l > 0
    BYTE[p] := (BYTE[p] & mask) | color
    --l
    p += DISPLAY_COLUMNS

PUB VerticalLine(column, row, len, color) | p, l, layer, r
{{
  Draw a vertical line.

  Arguments:
    column -- the column number for the vertical line, zero-based. The
              leftmost column is column 0. It is range-limited via masking
              to 0..MAX_DISPLAY_COLUMN (it probably should be limited via
              min/max operators instead.)
    row -- the row number for the top of the vertical line, zero-based, The
           topmost display row is row 0. It is range-limited via masking to
           0..MAX_DISPLAY_ROW
    len -- the length of the line in pixels.
    color -- the color to set the pixel. It may either be the RGB color or
             the display color

  Return Value:
    None.
}}
  column &= MAX_DISPLAY_COLUMN
  ' Draw the vertical line in two parts: the part in the top
  ' half, and the part in the bottom half.
  row &= MAX_DISPLAY_ROW
  len <#= DISPLAY_ROWS - row
  layer := 0
  color := ToDisplayColor(color)

  repeat ColorDepthBits
    if row < REFRESH_ROWS
      ' Line starts in the top half of the display.
      p := DisplayBufferHighLayerAddr + (row * DISPLAY_COLUMNS) + column + layer
      l := (REFRESH_ROWS - row) <# len 
      VerticalLineHelper(p, l, color & %111, %11_111_000)
      r := 0  ' continue with the first line of the bottom half
    else
      r := row - REFRESH_ROWS
      l := 0
    if len > l
      ' Bottom half
      p := DisplayBufferHighLayerAddr + (r * DISPLAY_COLUMNS) + column + layer
      VerticalLineHelper(p, len - l, (color & %111) << 3, %11_000_111)
    color >>= 3
    layer += DBUF_HI2LO_LAYER_DELTA
  
' ---------------------------------------
' TODO: Optimize for ColorDepthBits > 1
' But with the availability of the much faster PASM write helper, this is
' probably not a priority.
PUB Box(column, row, width, height, color) | r, c
{{
  Draw a box of a uniform color.

  Arguments:
    column -- the column number for upper-left corner, zero-based. The
              leftmost column is column 0. It is range-limited via masking
              to 0..MAX_DISPLAY_COLUMN (it probably should be limited via
              min/max operators instead.)
    row -- the row number for the upper-left corner, zero-based, The
           topmost display row is row 0. It is range-limited via masking to
           0..MAX_DISPLAY_ROW
    width -- the width, in columns, of the box.
    height -- the height, in rows, of the box.
    color -- the color for the box. It may either be the RGB color or
             the display color

  Return Value:
    None.
}}
  column &= MAX_DISPLAY_COLUMN
  width <#= DISPLAY_COLUMNS - column
  row &= MAX_DISPLAY_ROW
  height <#= DISPLAY_ROWS - row
  color := ToDisplayColor(color)
  ' Pick one of the following two versions.
  'repeat r from row to (row + height - 1)
  '  HorizontalLine(col, r, width, color)
  repeat c from column to (column + width - 1)
    VerticalLine(c, row, height, color)
  
' ---------------------------------------
PUB RomCharacter(ch, col, row, width, height, ltrim, ttrim, bgcolor, fgcolor) {
  } | chbase, pBuf, p, v, fgc, bgc, r
{{
  Draw a ROM Font character.

  The character is clipped to the display, so it may appear to "hang off" of
  the edge(s) of the display.

  It may also be clipped when displayed, by using ltrim and rtrim. The
  character bitmap is "trimmed" before it is placed, so the "row" and "col"
  position will show the upper-left _untrimmed_ pixel of the character.

  "Trimming" (ltrim and ttrim) is a way to "skip over" the blank rows and
  columns around the edges of many of the characters.

  Arguments:
    ch -- the Character code (ASCII) to draw.
    col -- the column number for upper-left corner, zero-based. The leftmost
           display column is column 0. It is clipped to the actual columns of
           the display, so a character can appear to "hang off" of the edges
           of the display.
    row -- the row number for the upper-left corner, zero-based, The
           topmost display row is row 0. It is clipped to the actual rows of
           the display so a character can appear to "hang off" of the edges
           of the display.
    width -- the width, in columns, of the display area for the character.
    height -- the height, in rows, of the display area for the character.
    ltrim -- the number of columns on the left of the character bitmap to
             skip over.
    ttrim -- the number of rows on the top of the character bitmap to skip over
    bgcolor -- the color to display for empty (zero) character pixels.
    fgcolor -- the color to display for non-empty (one) character pixels.  

  Return Value:
    None.
}}
  ' Characters are normally 16 pixels wide, but we will only show width columns
  ' and we will prune on the left or right at the display edges. We will also
  ' trim some columns from the bitmap displayed based on ltrim.
  ' Simiarly for rows, height, and ttrim. 

  ' Adjust for the top edge
  if row < 0
    ' Trim more of the top.
    height += row  ' same as height -= -row
    ttrim -= row  ' same as ttrim += -row
    row := 0
  ' Adjust for the left edge
  if col < 0
    ltrim -= col
    width += col
    col := 0
  if ((row + height) > DISPLAY_ROWS)
    height -= (row + height) - DISPLAY_ROWS
  if ((col + width) > DISPLAY_COLUMNS)
    width -= (col + width) - DISPLAY_COLUMNS

  ' See if there is anything left to display
  if (width =< 0) or (height =< 0) or (row => DISPLAY_ROWS) or (col => DISPLAY_COLUMNS)
    return

  ' ROM characters are interleaved, even and odd together.
  chbase := $8000 + ((ch & $FE) << 6) + (ttrim << 2)

  ' for each row, we may have to prune off some of the pixels, when
  ' width < 16. Since we have to reverse the bit order, we do the reversal
  ' with the expression "v ~> (width << 1)" instead of "v ~> 32"
  ' Then we also shift by one bit depending on whether it is an even or
  ' odd character.
  ' We also need to worry about where the row is: top half (0..31) or
  ' bottom half (32..63).
  ' The slow way is compute the byte address in the display buffer for
  ' each row; faster is to divide it into top and bottom ranges, like
  ' the vertical line logic.

  ' Do each row separately.
  pBuf := DisplayBufferHighLayerAddr + (row * DISPLAY_COLUMNS) + col
  repeat r from row to (row + height - 1)
    p := pBuf
    ' Extract just the range of bits between the ltrim and rtrim areas.
    v := LONG[chbase] >> ((ltrim << 1) + (ch & 1))
    'v := LONG[chbase] >> (ch & 1)
    fgc := fgcolor
    bgc := bgcolor

    repeat ColorDepthBits
      if r < REFRESH_ROWS
        ' character row is in top half of the display.
        HorizontalHelper(p, width, v, 2, %11_111_000, fgc & %111, bgc & 111)
      else
        ' character row is in bottom half of the display
        HorizontalHelper(p - BYTES_PER_COLOR_LAYER, width, v, 2, {
          }              %11_000_111, (fgc & %111) << 3, (bgc & %111) << 3)

      ' Advance the layer for this row
      p += DBUF_HI2LO_LAYER_DELTA
      fgc >>= 3
      bgc >>= 3
    
    ' Here we advance/reinit for the next row.
    pBuf += DISPLAY_COLUMNS
    chbase += 4

' ---------------------------------------
PUB Bitmap1Bpp(bitmap_ptr, bitmap_ptr_row_incr, col, row, width, height, {
  }            ltrim, ttrim, bgcolor, fgcolor) | pBuf, r, pb, v, fgc, bgc
{{
  Draw a 1 bit-per-pixel (1bpp) rectangular bitmap in an area of the display.

  Only a portion of the bitmap may be displayed: a number of rows and/or
  columns of the butmap may be skipped over using "ltrim" and "ttrim"

  The character is clipped to the display, so it may appear to "hang off" of
  the edge(s) of the display.

  This only handles bitmaps up to 32 columns (bits) wide.
  A bitmap look like a vector of values (BYTE, WORD, OR LONG). Each value is
  the bit pattern for the pixels of one row of the bitmap, topmost row first.
  The bits in each row value have the leftmost pixels in the loworder bits.
  The code can handle BYTE, WORD or LONG storage for a bitmap by examining the
  alignment of the bitmap_ptr, and uses the "bitmap_ptr_row_incr" to advance
  the correct distance to the next bitmap row value.

  Arguments:
    bitmap_ptr -- the (hub) address of the top row of the bitmap. This may be
                  BYTE, WORD or LONG aligned; if for example the pointer is for
                  an odd address it will be considered a BYTE-aligned bitmap
                  and only 8 bitmap bits will be available in that row.
    bitmap_ptr_row_incr -- the distance to advance the "bitmap_ptr" value to
                           move to the next row; typically the values 1, 2, or
                           4, but may be larger.
    col -- the column number for upper-left corner, zero-based. The leftmost
           display column is column 0. It is clipped to the actual columns of
           the display, so a bitmap can appear to "hang off" of the edges
           of the display.
    row -- the row number for the upper-left corner, zero-based, The
           topmost display row is row 0. It is clipped to the actual rows of
           the display so a bitmap can appear to "hang off" of the edges
           of the display.
    width -- the width, in columns, of the display area for the bitmap.
    height -- the height, in rows, of the display area for the bitmap.
    ltrim -- the number of columns on the left of the bitmap to skip over.
    ttrim -- the number of rows on the top of the bitmap to skip over
    bgcolor -- the color to display for empty (zero) pixels.
    fgcolor -- the color to display for non-empty (one) pixels.  

  Return Value:
    None.
}}
  ' NOTE: No skip-pattern support, as is available in the PASM-based helper.
  '
  ' The bitmap is shown on the rows row..(row+height-1), and col..(col+width-1).
  ' ttrim and ltrim determine how to align the portion of the bitmap to show on
  ' the display.

  ' Adjust for the top edge
  if row < 0
    ' Trim more of the top.
    height += row  ' same as height -= -row
    ttrim -= row  ' same as ttrim += -row
    row := 0
  ' Adjust for the left edge
  if col < 0
    ltrim -= col
    width += col
    col := 0
  ' Adjust for the bottom edge
  ' Adjust for the right edge
  if ((row + height) > DISPLAY_ROWS)
    height -= (row + height) - DISPLAY_ROWS
  if ((col + width) > DISPLAY_COLUMNS)
    width -= (col + width) - DISPLAY_COLUMNS
  'if ((row + height) > DISPLAY_ROWS)
  '  height -= DISPLAY_ROWS - (row + height)
  'if ((col + width) > DISPLAY_COLUMNS)
  '  width -= DISPLAY_COLUMNS - (col + width)

  ' See if there is anything left to display
  if (width =< 0) or (height =< 0) or (row => DISPLAY_ROWS) or (col => DISPLAY_COLUMNS)
    return

  ' Adjust the pointer to the row for the starting row
  ' For each row
  '   get the row bits, trimmed on the right
  '   For each color layer
  '     Draw the row with the fg and bg color
  '     advance the row ptr.
  bitmap_ptr += ttrim << 1  ' skip the trimmed rows, word-sized
  pBuf := DisplayBufferHighLayerAddr + (row * DISPLAY_COLUMNS) + col
  repeat r from row to (row + height - 1)
    pb := pBuf
    ' Extract just the range of bits between the ltrim and rtrim areas.
    if bitmap_ptr & 1
      v := BYTE[bitmap_ptr]
    elseif bitmap_ptr & 2
      v := WORD[bitmap_ptr]
    else
      v := LONG[bitmap_ptr]
    v >>= ltrim
    fgc := fgcolor
    bgc := bgcolor

    repeat ColorDepthBits
      if r < REFRESH_ROWS
        ' character row is in top half of the display.
        HorizontalHelper(pb, width, v, 1, %11_111_000, fgc & %111, bgc & 111)
      else
        ' character row is in bottom half of the display
        HorizontalHelper(pb - BYTES_PER_COLOR_LAYER, width, v, 1, {
          }              %11_000_111, (fgc & %111) << 3, (bgc & %111) << 3)

      ' Advance the layer for this row
      pb += DBUF_HI2LO_LAYER_DELTA
      fgc >>= 3
      bgc >>= 3
    
    ' Here we advance/reinit for the next row.
    pBuf += DISPLAY_COLUMNS
    bitmap_ptr += bitmap_ptr_row_incr

