{{
HUB75 panel driver obect, PASM display refresh cog using video output.

Author:         Tom Dinger (Ding-Batty)
Release Date:   2024/03/05

Copyright 2024 Thomas E. Dinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

--------------------------------------------------------------------------

This file contains the Display refresh code for HUB75 display panel(s).
It is written in PASM for sufficient speed to avoid flicker, using the
video generator to write the parallel bit streams, and CTRB to generate
the HUB75 shift clock.

This is mostly statically configured, using settings from the hub75_config.spin
file; the final initialization is deferred until after the cog is started,
so that the hub memory that contained the PASM code loaded into the cog can
be recovered and used for the display buffer, and other buffers.

For more information:
---------------------

See the file "hub75_info.txt" for some details on the display interface
called "HUB75", with some useful links.

Display Buffer
==============

This driver assumes the following memory organization:

Each BYTE of the display buffer contains 3 color bits for each of two RGB LEDs
in the same column, but REFRESH_ROWS apart. The six color bits are stored in
the low-order bits of the byte: %xx_RGB_rgb (for whichever color bit order is
configured). These are written to OUTA using the video generator in the cog
to write out the six bis using the VGA output mode.

The color values for one row are stored in consecutive bytes, with the lower-
address byte containing the lower columns. Rows are also stored consecutively,
with the rows at the top of the display at the lower addresses, so the LED
for row == 0 and column == 0 is the upper-left corner LED.

A Columns x Rows display will use (Columns + Rows) / 2 bytes of storage.
So, a 32 x 32 display will use 512 bytes of storage, and a 64 x 64 display
will use 2048 bytes of storage, for one color bit.

For color depth > 1 (more than 1 bit per color), the bits are stored in
"layers" with one layer for each color bit. These layers are also stored
consecutively, with the lowest-order color bit first (this can change.)

The memory layout for the color bits makes it easier to refresh the display,
but likely requires help with writing into the display buffer from a write-
helper cog, as Spin is too slow for writes of any large numbers of pixels.

Configuration and Initialization
================================

Most of this object is initialized "statically" from the configuration
constants. But two settings are received at runtime:
- The color bit depth, which is from 1 to 8 (I might be able to raise the upper
  limit for smaller panels).
- The base address of the display buffer.

The color bit depth is not a property of the hardware, but configured by the
application in the Init call and patched into the PASM cog before it is started.
So, the color bit depth cannot change for the run of the cog.

The base address of the display buffer is passed to the cog code indirectly
through the PAR register only after the cog has been started. The PAR register
points to a hub LONG that will contain the starting address of the display
buffer; the cog wait for it to become non-zero, and then uses that display
buffer address.

This initialization delay permits the application to load the COG, and then
lay out the memory buffers including the display buffer, and the app is able
to include the HUB memory contains the cog PASM code during the layout
decisions.

When the driver initializes, it configures the PLL for the video generator to
run at the same rate as the system clock (e.g. 80MHz), so that the clocks are
pretty well synchronized (with about a half-cycle, with some jitter). It then
configures the video generator to use the PLL clock, and use 4 PLL clocks per
video pixel, and 4 * 4 == 16 PLL clocks per frame (4 pairs of led bits).

The driver then measures the alignment of hub access against the "waitvid
hand-off point" (WHOP), and adjusts the relationship for "optimal" alignment,
as discussed some in the following thread:
https://forums.parallax.com/discussion/131254/fyi-waitvid-vs-hub-window

@kuroneko also published some demonstration video drivers usin this technique,
for example in:
https://forums.parallax.com/discussion/135844/poc-most-useless-piece-of-code-ever-the

(There is discussion about this in a number of other threads as well, but they
predate the latest forum migration and lots of links are broken, so they take
quite a bit of work to find.)

Once the hub access point (every 16 clocks) and the WHOP point that reloads the
next for pixels of data are alignemd properly, the code then walks over the
display buffer one LONG at a time and passes it to the video at the WHOP. For
each displayed row, it also generates the HUB75 shift clock signal by running
CTRB configured as an NCO for sysclk/4 (which will keep it aligned with the
hub accesses and WHOPs), and turns it on at the start of each row and off at the
end. This requires careful cycle-counting, as is noted in the comments.

The 16-clock repeated operation is:
                    rdlong  pvColors, pvHubFetchAddr        ' WHOP + 8, +12
                    cmp     pvColors, #PIXEL_ORDER          ' WHOP
                    add     pvHubFetchAddr, #4              ' WHOP + 4

Because this is exactly 16 clocks, the loop body has to be unrolled, which
limits the longest row length to the size that will fit into the cog memory.
However, we don't need to count cycles when not sending shift clocks, because
the hub access and WHOP are always aligned.

Performance
===========

In Spin only, we can get about 17 to 21 panel refreshes per second, with only
one-bit color depth for a 64 x 64 display. This results in a lot of flickering
on the display.

This version of the display driver uses the video generator to write the bits
for each pair of bytes in parallel, and CTRB is used to generate the
synchronized shift clock pulses to advance the six display shift registers.
The driver uses 4 sysclks per pixel pair, and for a pair of 64 x 64 panels
(128 x 64) and color bit depth of 3, it reaches >600 refreshes per second.
}}

OBJ
  ' Pick up the static configuration.
  Config: "hub75_config.spin"

CON
  ' Import the needed configuration settings from the Config object, for
  ' code brevity.

  REFRESH_ROWS         = Config#REFRESH_ROWS
  MAX_REFRESH_ROW      = Config#MAX_REFRESH_ROW
  DISPLAY_COLUMNS      = Config#DISPLAY_COLUMNS
  MAX_DISPLAY_COLUMN   = Config#MAX_DISPLAY_COLUMN
  SHIFT_CLOCK_PIN      = Config#SHIFT_CLOCK_PIN
  SHIFT_LATCH_PIN      = Config#SHIFT_LATCH_PIN
  OE_INVERTED_PIN      = Config#OE_INVERTED_PIN
  SHIFT_CLOCK_MASK     = Config#SHIFT_CLOCK_MASK
  SHIFT_LATCH_MASK     = Config#SHIFT_LATCH_MASK
  OE_INVERTED_MASK     = Config#OE_INVERTED_MASK
  ROW_NUMBER_MASK      = Config#ROW_NUMBER_MASK
  ROW_NUMBER_INCREMENT = Config#ROW_NUMBER_INCREMENT
  COLORS_BASE_PIN      = Config#COLORS_BASE_PIN
  COLORS_MASK          = Config#COLORS_MASK

  RED_BIT              = Config#RED_BIT
  GREEN_BIT            = Config#GREEN_BIT
  BLUE_BIT             = Config#BLUE_BIT

  REFRESH_ROWS_SHIFT   = >| MAX_REFRESH_ROW

  ' Define the order to output the bytes for the four pixels per long:
  ' low-order bits are first both for the pixel index (below) and the
  ' bytes in the LONG containing the pixel colors. This is used as the
  ' "Pixels" parameter for the WAITVID (or the equivalent video generator
  ' load operation), to select the four bit patterns for the color bits
  ' from the "Colors" WAITVID parameter.
  PIXEL_ORDER = %%3210

  ' We want to control the offset of the WHOP (and the appearance of
  ' the first output bit of the data provided in the WHOP) from the hub access
  ' point, so we can have predicatable timing behavior.
  ' So, we want to know a few things:
  ' - Where in the WAITVID is the WHOP?
  ' - Where in the WAITVID (or after) is the first bit available?
  ' We control the timing of the finish of the hub access and the following
  ' WAITVID completion, in the range of 4..20, and measure the timing of various
  ' elements.
  ' We also want to know if there is a "tolerance" of a range of clocks for
  ' picking up the two parameters in the WHOP.
  '
  ' It looks like the "sweet spot" for range of control is 7..10, which puts
  ' the WHOP on the instruction immediately after the hub access.
  ' And adjusting the WHOP by one clock _does_ in fact adjust its position
  ' relative to the CTRB signal, so we have some flexibility by adjusting
  ' both CTRB phase and WHOP_OFFSET.
  '
  ' WHOP_OFFSET == 9 and CTRB PHSB == $4000_0000 puts the rising edge of the clock
  ' in the middle of each video bit.

  WHOP_OFFSET = 9
      '  6 == pcZero
      '  7 == pcData1
      '  8 == pcData1
      '  9 == pcData1
      ' 10 == pcData1
      ' 11 == pcData2
      ' 12 == pcData2
      ' 13 == not clear (junk?)
      ' 14 == not clear (junk?)
      ' 15 == pcZero
      ' 16 == pcZero
      ' 17 == pcZero
      ' 18 == pcZero
      ' 19 == pcZero
      ' 20 == a few extra pulses

  ' For a WHOP_OFFSET adjustment of 9 above, we have the following "slots":
  '    +00       WHOP
  '    +04       Start or stop shift clocks in CTRB wth PSHB == $4000_0000
  '              by setting FRQB to $0100_00000 << 23
  '    +08, +12  Hub access aligned

  ' COLOR_PINS_VIDEO_GROUP is 0..3
  COLOR_PINS_VIDEO_GROUP = (COLORS_BASE_PIN >> 3) & 3

  ' Define the maximum number of columns we can handle, i.e. for which the
  ' unrolled column loop plus overhead will fit into the cog.
  ' For each additional 16 columns we need 12 free longs.
  MAX_COLUMNS = 37 * 16    ' == 592

  ' The number of loop instructions to copy == #pSend4Pixels - 1, or
  ' ((MAX_COLUMNS >> 2) * 3) - 4
  COPY_LOOP_COUNT = ((MAX_COLUMNS >> 2) * 3) - 4


'=========================================================================
' Validate the various configuration settings:
' - Pins are valid, and do not conflict with each other
' - display configuration "makes sense"
'
' The validation is done using FIT statements to make logical assertions
' about the config values, such as "is a power of two", "has one of a few
' specific values", etc.
' ------------------------------------------------------
' Verify the above settings.
'
' Now, use a "trick" to get the Spin compiler to check that the values for
' CACHE_LINE_LENGTH and CACHE_SIZE_BYTES are acceptable:
' - Make sure they both are powers of two.
' - Make sure they both are in the supported range
' - Make sure the pair is a supported combination.

DAT
  ' If the condition is false (==0) then the "org 1" means that the
  ' fit statement will generate an error. So we set the conditionals up
  ' so that a unsupported configurations generate a false (==0) value.

  org 1

  ' The REFRESH_ROWS must be a power of two.
  ' Error means REFRESH_ROWS is not a power of 2
  fit ||((REFRESH_ROWS == |< ((>| REFRESH_ROWS)-1)))

  ' The DISPLAY_COLUMNS must be a multiple of four, because we use four
  ' pixels per LONG of hub memory.
  ' Error means DISPLAY_COLUMNS is not a multiple of 4
  fit ||((DISPLAY_COLUMNS & 3) == 0)

  ' Make sure we can handle the number of columns as configured.
  fit ||(DISPLAY_COLUMNS =< MAX_COLUMNS)

  ' The COLORS_BASE_PIN must be one of 0, 8, 16, or 24, as the base pin of
  ' a (VGA) video pin group.
  ' Error means the COLORS_BASE_PIN was given an unspported value.
  ' For some reason FlexSpin doesn't like the "OR" operator:
  '   "error: syntax error, unexpected instruction"
'  fit ||((COLORS_BASE_PIN ==  0) OR (COLORS_BASE_PIN ==  8) OR {
'}        (COLORS_BASE_PIN == 16) OR (COLORS_BASE_PIN == 24))
  ' The following seems to work FlexProp, and the other compilers:
  fit ||((COLORS_BASE_PIN ==  0) + (COLORS_BASE_PIN ==  8) + {
}        (COLORS_BASE_PIN == 16) + (COLORS_BASE_PIN == 24))

  ' This version of the code assumes that the first color layer in the buffer
  ' is the least-significant color bit, so we make sure that the config file
  ' is set up correctly.
  fit ||((Config#DBUF_HI2LO_LAYER_DELTA & $80000000) <> 0)
  '
  ' For most-significant bit in the first layer, use:
  'fit (Config#DBUF_LO2HI_LAYER_DELTA & $80000000) <> 0

  ' We could do a lot more pin assignment collision testing. Caveat Emptor.

'=========================================================================

PUB NoOp
{{
  Every Spin Object must have at least one Spin PUB function, and the first
  one may be run automatically, so we make sure the first one is "safe."
  This one does nothing.

  Arguments:
    None

  Return Value:
    None
}}
  ' Do nothing

'=========================================================================
PUB LoadDisplayCog(ColorDepth, DisplayBufAddrVar, RefreshClocksAddrHub)
{{
  Add the runtime-configured parameters to the hub image for the cog PASM code
  before loading the cog:
  - Color Depth (bumber of bits in the each color, i.e. the number of "layers"
  - Address of the hub variable that will eventually get the address of the
    display buffer, to allow delayed initialization of the cog, so the
    application can reclaim the hub copy of the cog code.
  - The address of the location in the hub to get the count of sysclks used for
    a single full display refresh, so the refresh rate can be reported.
  Then the cog is started, which will load the cog, and then the cog code will
  wait until the display buffer address is supplied.

  Arguments:
    ColorDepth -- the numner of bits in each color value, which is the number
                  of color "layers" in the display buffer. E.g. "24-bit" color,
                  which is 8:8:8, would have a depth of 8.
    DisplayBufAddrVar -- the hub addres of the variable that will eventually
                         contain the actual hub address of the display buffer.
                         This can start out as zero, which will cause the
                         cog initialization code to stall, waiting for a non-
                         zero address to be supplied. This is a "two-stage"
                         initialization for the cog.
    RefreshClocksAddrHub -- the hub address of the LONG variable that will
                            receive the number of sysclks needed for one full
                            refresh of the display, so that the refresh rate
                            can be calculated and reported.

  Return Value:
    The Cog ID of the display refresh cog, plus one, so that it is non-zero.
}}
  ' Patch the color depth into the Cog code, and load the cog, passing the
  ' pointer to the place the cog will (eventually) get the display buffer
  ' address.
  '
  ' Returns the cog# + 1 of the cog started.

  ' Put the final few values into the cog.
  pcFullRefreshClocksAddrHub := RefreshClocksAddrHub
  pcColorDepthBits           := ColorDepth

  ' Start the cog, and pass the address in Hub memory that will eventually get
  ' the hub address of the display buffer. The cog, after starting, waits until
  ' that location contains a non-zero hub address, and uses that address as the
  ' start of the display buffer.
  return 1 + COGNEW(@pPasmStart, DisplayBufAddrVar)

' ========================================================================
'
' PASM code for display refresh
'
' Style notes:
' I use special prefixes for code labels, variables and constants stored in
' the pasm code, so that names make clear what is being referred to.
'
' "pCodeLabel" -- A label on code (possibly with a leading ":")
' "pvVariable" -- A label on a variable stored in a register in the cog,
'                 that can change over the course of cog execution
' "pcConstant" -- A label on a value that is effectively constant for the
'                 lifetime of the cog -- it may be set at initialization time,
'                 but will not be changed later.
' "piInstruction" -- A label on an instruction that may be the target of
'                    runtime instruction modification.

DAT
                    org 0
pPasmStart
                    jmp     #pInitContinue

' We store all the init-time constants and variables in low cog memory, so
' that we can recover them (and the init code) by expanding the unrolled
' column output loop body on top of them after init is complete.

' Hub addresses for the refresh clock count. Set by the initialization code
' that starts the cog. This is used during init of the cog, and the number
' of clocks for one full refresh is calculated and written there. It is
' not needed later. 
pcFullRefreshClocksAddrHub  long   0

' pcColorDepthBits is the number of bits per color. Set by the initialization
' code that starts the cog. This is only used during init of the cog.
pcColorDepthBits    long     3

' pcOutputLongs3x is the distance from the _end_ of the unrolled column
' output loop to the entry point, for the current display configuration.
' We have four display columns per long, and three instructions per
' unrolled loop body for each long. 
pcOutputLongs3x     long    (DISPLAY_COLUMNS >> 2) * 3

' Control pin mask definitions.

pcOEInvertedMask    long    OE_INVERTED_MASK

' pcDiraMask is the mask of output bits to set in DIRA when the cog starts.

pcDiraMask          long    COLORS_MASK | ROW_NUMBER_MASK | {
  }                         SHIFT_CLOCK_MASK | SHIFT_LATCH_MASK | {
  }                         OE_INVERTED_MASK

' Simple constants
pcZero               long    0
pc_4000_0000         long    $4000_0000  ' Used during init

'                        v: x -- don't care
'                        | vvvvv: ctr mode %00001: PLL, no output pins
'                        | ||||| vvv: PLLDIV for PLL / 1
pcCtraMode      LONG    %0_00001_111_00000000_000000_000_000000

'                         v: x -- don't care
'                         | vvvvv: ctr mode %00100: NCO, single-ended.
'                         | |||||              vvvvv: CTRB output A Pin
pcCtrbMode      LONG    (%0_00100_000 << 23) | SHIFT_CLOCK_PIN

'                        v: 4 pllclk per pixel
'                        |          vv: 4 pixels per "frame" == 4 * 4 pllclks
pcVidSCL        LONG    (4 << 12) | 16

'                        v: don't care
'                        | vv: Video Mode %01: VGA
'                        | || v: CMode: 4-color (2 bpp)
'                        | || | vv_vvv: Chroma0/1 + aural sub -- don't care.
'                        | || | || |||         vvv: Vgroup -- e.g. 1 for pins 15..8
'                        | || | || |||         |||                            vvvvvvvv: output pins in group
pcVidCFG        LONG    %0_01_1_00_000 << 23 | COLOR_PINS_VIDEO_GROUP << 9 | %00111111
pvDummy         LONG    0
pvDelay         LONG    0
pvTemp          LONG    0       ' temporary variable

' Stores the number of system clocks needed for one full refresh of all color
' layers for the display. This is copied back into hub memory so that the main
' program can report the performance of the driver, in refreshes per second.
pvFullRefreshClocks   LONG    0

' To stop the cog, the main program can also set the OEInverted pin for
' output, and can write a 1 to it to override anything this cog can do
' (output state is OR'd amoung all the cogs); then this cog can be
' stopped.

'-------------------------------------------------------------------------

' The code continues here...
pInitContinue
                    ' Wait for the app to pass the display buffer
                    ' address to us.
:pInitWaitLoop
                    rdlong  pcBufAddrHub, PAR  wz
    if_z            jmp     #:pInitWaitLoop

                    ' Calculate some helpful constants
                    mov     pcMaxLayerRowRepeatCountLess1, #1
                    shl     pcMaxLayerRowRepeatCountLess1, pcColorDepthBits
                    shr     pcMaxLayerRowRepeatCountLess1, #1
                    sub     pcMaxLayerRowRepeatCountLess1, #1

                    ' Set OEInverted so that the output starts disabled.
                    ' All other output bits are zero.
                    mov     OUTA, pcOEInvertedMask

                    ' We have some more one-time work to do:
                    ' - calculate the loop count for copying the loop body
                    '   over this init code;
                    ' - calculate the loop entry point, that is jumped to by
                    '   the start-of-row logic. (Done above)
                    ' In theory, we only have to copy enough loop body copies
                    ' to cover the number of output columns. TODO: See if that
                    ' is easy, vs. just copying the max number of copies.
                    '
                    ' The max number of loop instructions to copy is:
                    '      #pSend4Pixels - 1
                    ' or
                    '      ((MAX_COLUMNS >> 2) * 3) - 4
                    'mov     pvCopyLoopSize, #1  ' placeholder.
                    ' Now, statically initialized.

                    ' Calculate the entry point for the unrolled loop body,
                    ' using the number of output longs
                    mov     pcSendStartAddress, #pSend0Pixels
                    sub     pcSendStartAddress, pcOutputLongs3x
                    add     pcSendStartAddress, #1   ' Offset to the instruction to update
                    ' The +1 will be adjusted back -1 after the second instruction
                    ' of the unrolled loop is modified.

                    ' Do some start-of-refresh-pass initializations here, so that
                    ' they aren't needed once we start counting WHOP clocks.
                    mov     pvHubFetchAddr, pcBufAddrHub
                    mov     pvLayerRowRepeatCount, #1
                    mov     pvRowRepeatLoopCnt, pvLayerRowRepeatCount

                    ' Instead of measuring the number of clocks for a full
                    ' refresh of all layers, we can simply calculate that here,
                    ' because of the strict timing constraints on writing row
                    ' bits:
                    ' Each row write has 64 clocks (4 WHOP periods) of overhead, so
                    ' one row for one color bit takes
                    '    (8 * pcDisplayColumns) + 64 clocks
                    ' we have REFRESH_ROWS rows:
                    '    ((8 * pcDisplayColumns) + 64) * REFRESH_ROWS clocks
                    ' for color bits, we repeat each row
                    '    (|< pcColorDepthBits) - 1 times
                    ' There is also an extra WHOP period (16 clocks) between
                    ' each color bit layer.
                    mov     pvTemp, pcDisplayColumns
                    shl     pvTemp, #2   ' * 4
                    add     pvTemp, #64  ' + 64 == clocks for one row
                    shl     pvTemp, #REFRESH_ROWS_SHIFT ' * REFRESH_ROWS
                    mov     pvFullRefreshClocks, pvTemp
                    shl     pvFullRefreshClocks, pcColorDepthBits
                    sub     pvFullRefreshClocks, pvTemp
                    ' Add COLOR_DEPTH * 16
                    mov     pvTemp, pcColorDepthBits
                    shl     pvTemp, #4
                    add     pvFullRefreshClocks, pvTemp
                    wrlong  pvFullRefreshClocks, pcFullRefreshClocksAddrHub

                    ' Set up CTRA for PLL to generate sysclk/1
                    ' (This assumes FRQA and CTRA were already init'd to zero)
                    ' The FRQ divides sysclk by 16, the PLL multiplies it by 16,
                    ' and we use the PLLDIV of PLL/1.
                    movi    FRQA, #%0001_00000  ' 80MHz / 16 == 5 MHz
                                                ' vid PLL does * 16 back to 80MHz
                    ' We use PLL tap PLL / 1 to get 80 MHz, or 8 PLL clocks per
                    ' pixel == 8 sysclks per pixel.
                    mov     CTRA, pcCtraMode

                    ' Set up CTRB to generate a sysclk/4 signal in NCO mode.
                    ' We adjust the phase (by setting FRQB) when we know the
                    ' WHOP point, and we enable/disable it by setting or zeroing
                    ' the FRQB value.
                    ' We set the starting phase for the generated clocks here --
                    ' we turn the pulses on and off at the exact same point
                    ' within the pulses so the phase is always correct when we
                    ' turn it on again.
                    'movi    FRQB, #%0100_00000  ' 80MHz / 4 == 20 MHz
                    mov     FRQB, #0             ' Set later to the right value
                                                 ' for sysclk / 4
                    mov     PHSB, pc_4000_0000   ' Correct phase relative to the WHOP
                    mov     CTRB, pcCtrbMode

                    ' Set up video generator for 4 clock per pixel,
                    ' 16 clocks (4 pixels) per long, 2 bits per pixel.
                    ' (VGA mode).
                    mov     VSCL, pcVidSCL 
                    mov     VCFG, pcVidCFG

                    ' Wait 1 to 3 ms, to let the PLL settle down.
                    rdlong  CNT, #0   ' Get the current sysclk frequency
                                      ' Put it in the CNT shadow register
                    shr     CNT, #9   ' 1/512th of a second, about 2 ms
                    add     CNT, CNT  ' shadow-cnt has future CNT value
                    waitcnt CNT, #0   ' about 2 ms delay

                    ' Now, measure the sysclk relationship between the end of
                    ' WAITVID and the end of the next hub access. We then
                    ' "stretch" the VSCL frame length for one frame so that
                    ' the WHOP falls 5 to 8 sysclks after the end of the hub
                    ' access.
                    waitvid pcZero, #0
                    rdlong  pvDummy, #0     ' don't care about the data.
                    neg     pvTemp, CNT     ' the end of the hub access
                    waitvid pcZero, #0
                    ' We assume the WHOP is here, or before here.
                    add     pvTemp, CNT     ' now the delay amount

                    ' Based on the measurement, figure out how much to adjust
                    ' te frame size to delay the WHOP until it has the right
                    ' relationship with the hub access.
                    ' The WHOP_OFFSET value was calculated by trying different
                    ' values, and looking at the video pin output and the
                    ' shift-clock pin output on a dual-trace 'scope, until we
                    ' found the value that worked the best.
                    mov     pvDelay, #(WHOP_OFFSET + 32)
                    sub     pvDelay, pvTemp
                    and     pvDelay, #15

                    ' Stretch one frame
                    add     VSCL, pvDelay  ' one frame is stretched.
                    waitvid pcZero, #0 
                    sub     VSCL, pvDelay  ' next frame is back to normal
                    waitvid pcZero, #0
                    ' Here we are at WHOP + 8 (approximately)

                    ' We only need to keep track of the WHOP points during the
                    ' unrolled loop body, and getting into and out of the body,
                    ' because every hub access aligns with a WHOP, so we just
                    ' sync on the first hub read for a row.
                    '
                    ' However, we do "count clocks" so that we can determine the
                    ' (theoretical == actual) number of clocks for one full
                    ' refresh.

                    ' Ready to start refreshes.
                    mov     DIRA, pcDiraMask
                    jmp     #pFinishInit

' ------------------------------------------------------------------------
' Here we have the loop body, with padding, so that the loop body can take as
' much of the cog memory as possible, leaving only enough room for the start
' and end of row, layer, and refresh code and variables.
'
' So, the prologue to get into the loop body will look like this:
'                   rdlong  pvColors, pvHubFetchAddr      ' WHOP + 8, 12
'                   nop                                   ' WHOP  (data ignored)
'                   nop                                   ' WHOP + 4
'                   add     pvHubFetchAddr, #4            ' WHOP + 8
'                   ' NOTE: No "#" because we are jumping through the address
'                   ' stored there -- indirect jump.
'                   jmp     pvSendStartAddress            ' WHOP +12

                    long    0 [COPY_LOOP_COUNT - $]  ' Expand to leave enough room

' The unrolled row data loop body.
' The entry point is on a WHOP instruction, and we have to modify the next
' instruction to turn on the CTRB shift-clock generator with the right phase.

                    rdlong  pvColors, pvHubFetchAddr       ' WHOP + 8, 12
pSend4Pixels
                    cmp     pvColors, #PIXEL_ORDER         ' WHOP
                    ' Strictly speaking, if we are repeating the row, we don't
                    ' really need this pointer advance -- we can just subtract
                    ' less to go back ot the row start.
                    ' TODO: See if moving this make it easier to do other
                    ' optimizations.
                    add     pvHubFetchAddr, #4             ' WHOP + 4

                    ' The next set of steps turns off CTRB

                    ' Latch the row data, set the row number, enable output,
                    ' and advance the row number
                    ' We XOR-change the row number bits in OUTA to get an
                    ' "atomic" update of the row number without (briefly)
                    ' setting the row number to zero. My assumption is that
                    ' the row number is _not_ latched in any way.

                    ' This version triggers the ShiftLatch and disables output
                    ' in the same instruction that the row bits are changed,
                    ' which should not cause any ghosting. The second latch,
                    ' associated with the Output Enable pin, will not collect
                    ' the row settings until OEInverted is set to zero in the
                    ' next instruction.
                    ' (I suppose it is possible that OEInverted is slower to
                    ' propagate than the row number bits, but that should only
                    ' be a very few ns (or less time) of wrong-row display.)

                    xor     pvPrevRow, pvCurrentRow          ' WHOP + 8

' pSend0Pixels is the base address for calculating where to enter the loop
' based on the number of longs to output.
' The start of the unrolled loop body is pSend0Pixels - (3 * (DISPLAY_COLUMNS >> 2))
pSend0Pixels

                    mov     pvPrevRowTemp, pvPrevRow         ' WHOP +12
                    mov     pvPrevRow, pvCurrentRow          ' WHOP (data ignored)
                    mov     FRQB, #0                         ' WHOP + 4 ' end shift-clocks
                    xor     OUTA, pvPrevRowTemp              ' WHOP + 8 ' latch data, change row#, disable output
                    andn    OUTA, pcOEInvertedShiftLatchMask ' WHOP +12 ' enable output
                    djnz    pvRowRepeatLoopCnt, #pRepeatRow  ' WHOP (data ignored)
                    ' no jump takes up WHOP + 4

                    ' Here we advance to the next row, and see if we are done
                    ' with the layer (row number wrapped around).

                    add     pvCurrentRow, pcRowNumberIncr   ' WHOP + 8
                    and     pvCurrentRow, pcRowNumberMask wz 'WHOP +12 ' did we wrap around?

                    ' The following needs to go _somewhere_ after we change
                    ' pvPrevRow, above, but it can really go anywhere.
                    or      pvPrevRow, pcOEInvertedShiftLatchMask  ' WHOP (data ignored)

                    ' If Z is clear, then we are proceeding to the next row,
                    ' for which pvHubFetchAddr is correct for the next row start, but
                    ' we need to reset the row repeat loop count.
                    ' If Z is set, we have finished all the rows for the current
                    ' color bit, so we advance to the next layer.
    if_z            jmp     #pEndOfLayer                  ' WHOP + 4

pNextRowStart
                    ' Here we start the next row for the current layer.
                    rdlong  pvColors, pvHubFetchAddr      ' WHOP + 8, 12
                    mov     pvRowRepeatLoopCnt, pvLayerRowRepeatCount 'WHOP (data ignored)
                    jmp     #pNextRowStart2               ' WHOP + 4

                    ' Code can be inserted here to eliminate a jump and save
                    ' a long.

pNextRowStart2
                    add     pvHubFetchAddr, #4            ' WHOP + 8
                    ' NOTE: No "#" because we are jumping through the address
                    ' stored there -- indirect jump.
                    jmp     pcSendStartAddress            ' WHOP +12

' -------------------------
pEndOfLayer
                    ' All the rows for this color bit have been shown.
                    ' After each color bit, we adjust the row repeat count
                    ' (doubling it) so the next bit is shown for twice as long,
                    ' if not starting a new refresh pass.

                    ' See if we have just finished the most significant bit of
                    ' the color depth:
                    ' - As a special case: if COLOR_DEPTH = 2, we have
                    '   pcMaxLayerRowRepeatCountLess1 == 1, which is also
                    '   the repeat count for the first of the two layers, so
                    '   the CMPSUB will set the C bit and set the repeat count
                    '   for the next row to zero, which is not what we want.
                    '   Detect this by also checking for Z -- if C is set _and_
                    '   Z is set, this is _not_ the final row, and we need to
                    '   "fix" the pvLayerRowRepeatCount by setting it back to 1
                    '   before it is doubled for the next row.
                    '   This is the only case for which Z will be set.
                    ' - If we have finished the highest bit (the largest repeat
                    '   count), the "cmpsub" instruction will set the C flag, and
                    '   change the repeat count back to 1.
                    ' - If we have not finished the highest bit, then the
                    '   repeat count will be unchanges and C clear, and we should
                    '   shift the count left one bit and do the next layer.
                    cmpsub  pvLayerRowRepeatCount, pcMaxLayerRowRepeatCountLess1  wc, wz ' WHOP + 8
     if_nc_or_z     jmp     #pColorBitLayerStart         ' WHOP + 12

                    ' Start the next refresh pass
                    ' Here on WHOP (data ignored)

' This section, if enabled, measures the actual sysclks for one refresh, but
' it takes an extra 16 sysclks (an extra WHOP time) to accomplish. It is
' available for validating the predicted refresh clock calculated earlier.
' It also assumes that pvRefreshPassStartCnt was set just before the very first
' pass during late init, which will add another instruction there as well. So
' this all will reduce the MAX_COLUMNS by about 8 columns.
'                    mov     pvFullRefreshClocks, pvRefreshPassStartCnt ' WHOP (data ignored)
'                    neg     pvRefreshPassStartCnt, CNT                 ' WHOP + 4
'                    add     pvFullRefreshClocks, CNT                   ' WHOP + 8
'                    sub     pvRefreshPassStartCnt, #4                  ' WHOP +12

'                    ' Copy back to hub
'                    ' nop   ' hub delay                                ' WHOP (data ignored)
'                    ' nop   ' hub delay                                ' WHOP + 4
'                    wrlong  pvFullRefreshClocks, pcFullRefreshClocksAddrHub ' WHOP + 8, 12

                    mov     pvHubFetchAddr, pcBufAddrHub ' WHOP (data ignored)
                    jmp     #pNextRowStart               ' WHOP + 4

' -------------------------
pColorBitLayerStart                      
                    ' Fix the special case of COLOR_DEPTH=1 for the low-order
                    ' color layer by restoring the count to 1 again which was
                    ' changed to zero if Z is set.
    if_c_and_z      mov     pvLayerRowRepeatCount, #1    ' WHOP (data ignored)
                    shl     pvLayerRowRepeatCount, #1    ' WHOP + 4
                    ' Worst-case alignment here... Maybe we can fix that?
                    'nop     ' can be omitted             ' WHOP + 8
                    'nop     ' can be omitted             ' WHOP +12
                    'nop     ' can be omitted             ' WHOP (data ignored)
                    jmp     #pNextRowStart               ' WHOP + 4

' -------------------------
pRepeatRow
                    sub     pvHubFetchAddr, pcDisplayColumns 'WHOP + 4

                    rdlong  pvColors, pvHubFetchAddr      ' WHOP + 8, 12
                    nop     ' cannot be omitted           ' WHOP  (data ignored)
                    jmp     #pNextRowStart2               ' WHOP + 4

' ------------------------------------------------------------------------
pFinishInit

' The "final" init code.

' This expands the unrolled column loop body to replace the init code that
' we no longer need, and modify one instruction so that the CTRB shift-clock
' is turned on at the right time. Then it enters the row refresh loop.
'
' The number of loop instructions to copy == #pSend4Pixels - 1, or
' ((MAX_COLUMNS >> 2) * 3) - 4

pCopyLoopBodyLoop
                    mov      pSend4Pixels - 2, pSend4Pixels + 1
                    sub      pCopyLoopBodyLoop, pcIncSrcDest
                    djnz     pvCopyLoopSize, #pCopyLoopBodyLoop

                    ' Setup for this was done early in init.
                    movd    piFixCtrbInit, pcSendStartAddress
                    ' Two instructions moved down to fill the two nops, and
                    ' save two longs.
'                    sub     pcSendStartAddress, #1
'piFixCtrbInit       mov     0+0, pcStartCtrbInstruction

                    rdlong  pvColors, pvHubFetchAddr      ' WHOP + 8, 12
                    'nop     ' cannot be omitted           ' WHOP  (data ignored)
                    'nop     ' cannot be omitted           ' WHOP + 4
                    sub     pcSendStartAddress, #1
piFixCtrbInit       mov     0+0, pcStartCtrbInstruction
                    add     pvHubFetchAddr, #4            ' WHOP + 8
                    ' NOTE: No "#" because we are jumping through the address
                    ' stored there -- indirect jump.
                    jmp     pcSendStartAddress            ' WHOP +12

' ------------------------------------------------------------------------
' The compile-time initialized values, that are needed after initialization,
' to continually refresh the display.

' pcBufAddrHub contains the hub address of the first row and column of the
' lowest-significance color bit of the display buffer.
' Set by the initialization code that starts the cog.

pcBufAddrHub        long    0    ' Starting address of the entire buffer

' pcDisplayColumns is the number of display columns per row; this is _not_ the
'                  number of longs per row, and might not even be the number of
'                  _bytes_ per row, depending on how the data is packed.
'
' It is unlikely, but possible, for these to exceed 511 -- for example, 8
' panels in a series, each of which has 64 columns, would reach 512 columns.
' If we need cog space, we could get rid of them and limit the column count.

pcDisplayColumns    long    DISPLAY_COLUMNS

' pcOEInvertedShiftLatchMask is used during row number update to eliminate
' a runtime instruction. We can treat the shift latch and Output enabled
' as a single bit (for the ICN2037 controller chip, at least), because th
' two signals are inverted w.r.t. each other, and trigger on opposite edges. 

pcOEInvertedShiftLatchMask long OE_INVERTED_MASK | SHIFT_LATCH_MASK

' Instead of using a normal counter for the current display row number being
' updated, and shifting it into position to write it to the display, we
' manage the row number value in its already-shifted position, to save a few
' instructions in the row processing loop.

pcRowNumberMask     long    ROW_NUMBER_MASK
pcRowNumberIncr     long    ROW_NUMBER_INCREMENT

' The number of times to repeat display of each row, to get the right
' brightness for each layer.
pcMaxLayerRowRepeatCountLess1  long    0   ' (|< (pcColorDepthBits - 1)) - 1  ' runtime

' The following variables can use static initialization, which saves one runtime
' instruction to set them; otherwise they could use RES.
'pvColumnLoopCount     LONG    DISPLAY_COLUMNS >> 2
pvLayerRowRepeatCount LONG    1  ' runtime
pvRowRepeatLoopCnt    LONG    1
pvPrevRow             LONG    0
pvCurrentRow          LONG    0

' The following variables are only used at the end of init when growning the
' unrolled column loop body. So they are overlaid with some other variables
' only used during actual display refresh.
' Used for copying the unrolled column loop body.

pvColors              'res     1  ' Gets the current row/col color bits
pcIncSrcDest          LONG    1 << 9 | 1

pvCopyLoopSize        LONG    COPY_LOOP_COUNT

pvPrevRowTemp         'res     1
pcStartCtrbInstruction   movi     FRQB, #%0100_00000         ' WHOP + 4 ' start shift-clocks

' pcSendStartAddress will contain the first instruction of the unrolled
' column output loop (after the row prologue). This is used for every row
' refresh.
pcSendStartAddress    res     1
pvHubFetchAddr        res     1  ' The next BYTE to read

'====================================================================

PasmEnd             fit  $1F0

