{{
  Initialization code for measure_refresh_rate.spin

Author:         Tom Dinger (Ding-Batty)
Release Date:   2024/02/26

Copyright 2024 Thomas Dinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

--------------------------------------------------------------------------

This file contains init code for the test program, that can be recovered
later for additional buffer space.

}}

OBJ
  ' Pick up the static configuration for the display(s).
  Config : "hub75_config.spin"

  ' The following object will be "reclaimed" after initial use, along with
  ' most of this object.
  SerPasm    : "Db_FullDuplexSerial_pasm"  ' uses a cog

CON
  { Configuration for serial connection. }
  BAUDRATE = 115_200
  RXPIN    = 31    ' The usual pins.
  TXPIN    = 30

  ' Import some CONstants
  DISPLAY_COLUMNS = Config#DISPLAY_COLUMNS
  REFRESH_ROWS    = Config#REFRESH_ROWS

  ' This is the distance from the last local variable address in the
  ' function that fills main memory that is safe to fill with the
  ' "unlikely" value.
  OFFSET_FOR_MAIN_STACK_FILL = $0010
  END_OF_STACK = $0000_8000      ' Highest hub RAM address
  UNLIKELY_VALUE = $DEADBEEF

  ' The initial amount of reserved memory, before the rest of the program
  ' that will be considered "available".
  ' This value works with BST, with unused function removal.
  'RESERVED_SIZE_BYTES = (1024 * 11) + (3900 * 4)
  ' This value works with PropTool
  RESERVED_SIZE_BYTES = (1024 * 11) + (3800 * 4)
  ' The P1-only Propeller Tool does not have that feature, so the number will
  ' likely need to be smaller.

  GLOBAL_VBASE_ADDR = $0008  ' WORD variable

DAT
  ' NextFree contains the lowest address currently available for a buffer.
  ' The maximum size available is LONG[GLOBAL_VBASE_ADDR] - NextFree bytes.
  ' When a buffer is "allocated" the value of NextFree is increased by the
  ' buffer size, rounded up to a long boudary.
  NextFree       LONG 0  ' @FreeMemStart

  ' This is the actual start of available memory. We need a real allocation
  ' here, in order to push the VBASE value as high in memory as we can.
  FreeMemStart   LONG 0 [(RESERVED_SIZE_BYTES + 3) >> 2]

' ========================================================================

PUB Init(pSerialCogId, pBuffers, pSerialConfig, pDisplayBufAddr, pRefreshClocks) {
      } | needed, max_color_depth
{{
  Perform all the "early" initialization operations for the program, so that
  this code can be recovered/reused for runtime buffers.

  The main program and this object _must_ agree on the layout and use of
  the buffers whose pointers are passed in. BEWARE: It is easy for these
  to get out of sync between this code and the main program.

  Arguments:
    pSerialCogIds -- points to the WORD location to get the Cog ID + 1 of the
               FDS PASM cog.
    pBuffers -- a list of two WORD values, that get the address and size
                reserved for the display buffer:
                  pBuffers + 0 -- the size in bytes of the Display Buffer
                  pBuffers + 2 -- the reserved address of the Display Buffer
                The logic here tries to allocate a display buffer with the
                largest COLOR_DEPTH possible, and returns that size and buffer.
                If the address entry is zero on return, then there
                was not enough available memory to satisfy the amount needed
                for the current configuration, even for COLOR_DEPTH=1
    pSerialConfig -- the configuration structure used to initialize the
                     FullDuplexSerial PASM cog, and used by the Spin routines
                     that read and write bytes to the serial connection.
                     The rx and tx buffer addresses and sizes are _not_ filled
                     in here.
    pDisplayBufAddr -- The address of the location in hub memory that the display
                       cog and the helper cog will watch for the actual display
                       buffer address, so that they will continue init.
    pRefreshClocks -- The address of the location for the display cog to put
                      its calculated or measured count of system clocks to
                      complete one full display refresh for all color bits, so
                      that the refresh rate can be calculated.

    Return value:
      == 0 -- buffer memory reservation/allocation succeeded.
      <> 0 -- not enough available memory for our buffer, the value returned
              is the number of bytes additional that is needed.
}}
  waitcnt(clkfreq * 4 + cnt)   ' Wait 4 seconds at startup, to allow time to
                               ' change focus to the serial window.
  ' Fill the main stack area so we can determine how much was used during
  ' runtime.
  FillMainStack
  WORDFILL(pBuffers, 0, 2)

  ' Since we are not using "dynamic" allocation of the serial buffers, we
  ' can start the serial cog now, because the caller can use the FDS object
  ' to report any problems with the rest of init.
  WORD[pSerialCogId] := SerPasm.LoadCog(RXPIN, TXPIN, %0000, BAUDRATE, SerPasm#NO_CTS, {
                                  }     pSerialConfig)
  ' If that failed, not much we can do here. (Although we could just hang...)

  ' Prepare to reserve our (large) buffers
  if NextFree == 0
    ' First time initialization.
    NextFree := @FreeMemStart

  ' Figure out the largest buffer we can allocate
  max_color_depth := (AvailableBytes / constant(DISPLAY_COLUMNS * REFRESH_ROWS)) <# 8
  if max_color_depth == 0
    ' Not enough memory for even one color layer
    return constant(DISPLAY_COLUMNS * REFRESH_ROWS) - AvailableBytes

  ' Do the allocation, which should not fail.
  WORD[pBuffers + 0] := max_color_depth * constant(DISPLAY_COLUMNS * REFRESH_ROWS)
  WORD[pBuffers + 2] := Allocate(WORD[pBuffers + 0])

  return 0

' ========================================================================
' Memory reservation/allocation routines.

PUB Allocate(size_in_bytes) | ptr
{{
  Try to reserve/allocate a buffer of the requested size.

  Arguments:
    size_in_bytes -- the size of the desired buffer.

  Return value:
    <> 0 -- the starting address of the reserved buffer; there was enough
            memory to satisfy an allocation of this amount.
    == 0 -- there is not enough memory to satisfy the allocation.
            The caller can use AdditionalNeeded(size_in_bytes) to find out
            the shortfall.
}}
  ptr := AdditionalNeeded(size_in_bytes)
  if ptr > 0
    return 0   ' not enough space
  ' There is enough room -- get the pointer, and advance NextFree
  ptr := NextFree
  NextFree += (size_in_bytes + 3) & constant(!$3)
  return ptr

' ------------------------------------------------------------------------
PUB AdditionalNeeded(size_in_bytes) | available
{{
  Calculate how much more reserved memory is needed to satisfy an allocation
  request for some bytes of memory.

  If there is not enough memory, this returns the (positive) number of
  additional bytes needed. This value should be added to RESERVED_SIZE_BYTES
  to get the needed value for RESERVED_SIZE_BYTES (which may not be possible
  without shrinking the main stack too much).

  Arguments:
    size_in_bytes -- the size of the desired buffer.

  Return value:
    0   -- there is enough memory to satisfy an allocation of this amount.
    > 0 -- there is not enough memory to satisfy the allocation. The value
           is the number of bytes of additional memory needed (
           "shortfall").

  The space available, in bytes, goes from NextFree to VBASE, i.e.
    WORD[GLOBAL_VBASE_ADDR] - NextFree
}}
  available := AvailableBytes
  ' Round size_in_bytes up to a multiple of LONGs
  size_in_bytes := (size_in_bytes + 3) & constant(!$3)
  if available => size_in_bytes
    return 0  ' Enough room
  return size_in_bytes - available  ' how much more room is needed.

' ------------------------------------------------------------------------
PUB AvailableBytes
{{
  Returns the number of bytes available for allocation/reservation.
}}
  if NextFree == 0
    ' First time initialization.
    NextFree := @FreeMemStart
  return WORD[GLOBAL_VBASE_ADDR] - NextFree

' ------------------------------------------------------------------------
PUB GetNextFree
{{
  Returns the hub address that would be the starting address of the next
  buffer to be allocated/reserved.

  Arguments:
    None

  Return value:
    The base addres that is the start of the next allocation.
}}
  return NextFree

' ========================================================================
PUB FillMainStack | p
{{
  Fills almost all of the unused main stack area with an "unlikely" value;
  When the program ends, that area can be examined to see how much was used
  and how much was not used.

  Arguments:
    None

  Return value:
    None
}}
  ' The OFFSET_FOR_MAIN_STACK_FILL is some extra distance into the stack to
  ' start writing, so that the LONGFILL does not overwrite any short-term
  ' stack it might be using.
  p := @p + OFFSET_FOR_MAIN_STACK_FILL
  LONGFILL(p, UNLIKELY_VALUE, (END_OF_STACK - p) >> 2)