{{
Test display and scrolling of digit columns in various ways, in preparation for
actual use for other purposes. Uses the HUB75 display.

Author:         Tom Dinger (Ding-Batty)
Release Date:   2024/02/22

Copyright 2024 Thomas E. Dinger

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

--------------------------------------------------------------------------

This is a test harness for testing various ways to scroll and update a
display made of digits, e.g. a counter, countdown clock, or time-of-day
clock.

This is used to determine the time it takes for various display update
approaches, to see which ones can be done within various time windows,
e.g. update a clock display every second. This depends on how complex the
logic to update a digit.

For example, if the display is divided into three bands with different
colors, and the digit update involves scrolling individual columns up and
down at the same time, there will be 26 * 6 19x26 bitmap writes to the display
per digit per second to get per-second updates. That may be possible, but
we should verify that (also, to get the proper delays to make the display
update "pretty").

[I'd like to discover that a six-digit display could be updated once a second
in the most complex display update method, so that a clock could be
implemented using that update approach. If only four digits can work that
way, we can use that for hours and minutes, and do something else for seconds,
such as using smaller digits, or simpler update for seconds, etc.]

This version should be implemented to adapt to the hub75_display_config.spin
settings for number of rows and columns.

Ways to show and update digits:
- single row of digits:
  - simple digit replace (fast replace) -- SimpleDigitsReplace()
  - slow digit replace, next digit replaces the previous one line at a
    time -- SlowDigitsReplace()
  - fade to black old digit, fade in new. (LATER)
  - fade to white (gray?) old digit, fade in new. (LATER)
- columns of digits, uniform color/intensity
  - scroll down from the top
  - scroll up from the bottom (same timing as scroll down)
  - interleaved scroll of pixel columns in the digit, half scroll up, half
    scroll down
- columns of digits, bold/different color for the "middle" digits.
  - scroll down from the top
  - scroll up from the bottom
  - interleaved scroll of pixel columns
    - single color during scroll, intensity/color change at end.
    - maintain three "layers" of colors.

How to measure time for updates:
- We need to use the feature of the helper opreations to get an indication that
  an operation is done, which we do for the final digit update.
- Advance all digits once, measure Sysclks from enqueuing the first write to the
  final write completing.
- Advance all the digits N times, measure the Sysclks a similar way:
  - They can all use the same completion status variable, as long as we only
    check it after the final command has been started/completed, so when the
    command FiFo is empty. (Needs a helper in the hub75_helper.spin code.) 
    - This is a problem -- instead, use a helper function that does a NoOp
      and waits for "done".
- Sysclks can be converted into an update rate per second, similarly to the
  calculation for refreshes per second for the display code.
  
}}
' Define the operating constants:

CON { General constants }
  _CLKMODE = XTAL1 + PLL16X 
  _XINFREQ = 5_000_000

  { Configuration for test serial connection. }
  BAUDRATE = 115_200
  RXPIN    = 31    ' The usual pins.
  TXPIN    = 30
  RX_BUFFER_SIZE = 28
  TX_BUFFER_SIZE = 200

  CR = $0D
  LF = $0A

OBJ
  ' Pick up the static display configuration first.
  Config: "hub75_config.spin"

CON
  { LED panel(s) configuration }
  ' Color depth of 2 through 6 works fine at 64 x 64
  COLOR_DEPTH   =  3   ' number of bits of resolution per color.
  MAX_COLOR_VALUE = (|< COLOR_DEPTH) - 1

  DISPLAY_ROWS       = Config#DISPLAY_ROWS
  REFRESH_ROWS       = Config#REFRESH_ROWS
  DISPLAY_COLUMNS    = Config#DISPLAY_COLUMNS
  MAX_DISPLAY_ROW    = Config#MAX_DISPLAY_ROW
  MAX_DISPLAY_COLUMN = Config#MAX_DISPLAY_COLUMN

  ' We extract the digits from the ROM Font. They are 16 x 32, but some of the
  ' rows and columns are blank, so we only keep 13 x 26 -- every digit (and
  ' the capital letter "O", used for zero) fits into 13-pixels wide, and
  ' using 26 rows is enough for a few blank pixel rows between rows of
  ' digits.
  { Digit "shapes" }
  DIGIT_WIDTH  = 13
  DIGIT_HEIGHT = 26

  ' We always have a row of digits centered vertially on the display.
  CENTER_DIGIT_TOP_ROW = (DISPLAY_ROWS - DIGIT_HEIGHT) / 2
  TOP_DIGIT_ROW_HEIGHT = CENTER_DIGIT_TOP_ROW
  BOTTOM_DIGIT_ROW_HEIGHT = DISPLAY_ROWS - DIGIT_HEIGHT - TOP_DIGIT_ROW_HEIGHT

  ' We always start the first column of digits in display column zero
  LEFT_DIGIT_LEFT_COLUMN = 0
  COLUMNS_PER_DIGIT = 16  ' four across a 64 x 64, eight across two panels
  MAX_DIGIT_COLUMNS = DISPLAY_COLUMNS / COLUMNS_PER_DIGIT

  ' For alternating pixel columns, we need a pair of bitmap masks to skip
  ' every other bit.
  SCROLL_UP_SKIP_MASK   = %0101_0101_0101_0101_0101_0101_0101_0101
  SCROLL_DOWN_SKIP_MASK = ! SCROLL_UP_SKIP_MASK

  ' We use a large command FiFo, because we may want to enqueue quite a
  ' few commands at times. Things really depend on how fast the helper
  ' cog can process the commands. We could always get a speed-up by using
  ' more than one display writer cog, because they can write to the
  ' display buffer in parallel. That is an optimization for another day?
  COMMAND_FIFO_SIZE = 32

OBJ
  Ser    : "Db_FullDuplexSerial"
  Umath  : "umath2.spin"
  Fifo   : "FifoBig.spin"
  Panel  : "hub75_display.spin"
  Helper : "hub75_helper.spin"

  ' Bring in the PASM components at the end, so we can (eventually) reclaim the
  ' hub memory to be used for buffers.
  SerPasm    : "Db_FullDuplexSerial_pasm"  ' uses a cog
  HelperPasm : "hub75_helper_pasm.spin"

  ' Pick exactly one of these to do the display refreshes.
  'PanelPasm  : "hub75_display_pasm_repeat.spin"
  PanelPasm  : "hub75_display_pasm_video_4clock.spin"
  'PanelPasm  : "hub75_display_pasm_video_8clock.spin"
  'PanelPasm  : "hub75_display_pasm_waitcnt.spin"

DAT
' Define the variables that contain the operating parameters.
' We use a DAT section so we can initialize them, which means there can be only
' one instance of this object; if we used a VAR section, we'd have to devote
' code to initializing the VARs, which might happen later, depending on how the
' code develops.

' The storage for the command Fifo entries, used in the helper object to
' communicate with the helper PASM cog.
' We could allocate this any number of ways: VAR, DAT, "dynamically" like the
' display buffer.
CmdFifoData      LONG    0 [Fifo#HEADER_SIZE_LONGS * COMMAND_FIFO_SIZE]

' Storage for debug output from the helper driver
' See that driver for details on the contents.
' Should be at least 10 longs
HelperDebugOutput LONG    0[16]

' Color values used in the tests.
' This "knowns" the COLOR_DEPTH
BoldColorRgb      LONG    (MAX_COLOR_VALUE << 16) | (MAX_COLOR_VALUE << 8) | 0  ' bright yellow
BoldColorDisplay  LONG    0  ' Filled in at runtime.

DimColorRgb       LONG    (0 << 16) | (1 << 8) | 0  ' dim green
DimColorDisplay   LONG    0  ' Filled in at runtime

' Storage for the actual digit values being displayed.
DigitValues       BYTE    0[MAX_DIGIT_COLUMNS]
NewDigitValues    BYTE    0[MAX_DIGIT_COLUMNS]

DigitBitmapRow    WORD    0[MAX_DIGIT_COLUMNS]
NewDigitBitmapRow WORD    0[MAX_DIGIT_COLUMNS]
' The second starting position, for digits scrolling up
DigitBitmapRowUp  WORD    0[MAX_DIGIT_COLUMNS]

' For the display laypouts that uses three digit rows/regions, top, middle, and
' bottom, we store the layout info here:
top_row           LONG    0
top_row_height    LONG    TOP_DIGIT_ROW_HEIGHT
middle_row        LONG    CENTER_DIGIT_TOP_ROW
middle_row_height LONG    DIGIT_HEIGHT
middle_row_bitmap_offset LONG  TOP_DIGIT_ROW_HEIGHT << 1
bottom_row        LONG    CENTER_DIGIT_TOP_ROW + DIGIT_HEIGHT
bottom_row_height LONG    BOTTOM_DIGIT_ROW_HEIGHT
bottom_row_bitmap_offset LONG (CENTER_DIGIT_TOP_ROW + DIGIT_HEIGHT) << 1

' We may have more storage for other display tests. We use
' fixed storage and not locals because it may be more representative
' of the eventual implementations, that will need "too many" local variables.

' Digit storage
'
' For the scrolling digit tests, we treat the entire goupr of columns for one
' digit column as a large bitmap, so we can just change the starting row in
' the bitmap to update the entire column.
'
' For columns in which the digits descend from the top and exit toward the
' bottom, we actually want higher digits at lower addresses. For digit columns
' that scroll upward, higher digits are at higher addresses.
'
' We also need extra "guard digits" at both ends, a kind of overlap, so that
' the starting row for display can wrap backwards seamlessly, rather than
' using more complex code logic to glue the ends of the digit cycle together.

' Digits that scroll down.
DigitsDown
DigitsDown_1top      WORD    0[TOP_DIGIT_ROW_HEIGHT]
DigitsDown_0top      WORD    0[DIGIT_HEIGHT]
DigitsDown_9         WORD    0[DIGIT_HEIGHT]
DigitsDown_8         WORD    0[DIGIT_HEIGHT]
DigitsDown_7         WORD    0[DIGIT_HEIGHT]
DigitsDown_6         WORD    0[DIGIT_HEIGHT]
DigitsDown_5         WORD    0[DIGIT_HEIGHT]
DigitsDown_4         WORD    0[DIGIT_HEIGHT]
DigitsDown_3         WORD    0[DIGIT_HEIGHT]
DigitsDown_2         WORD    0[DIGIT_HEIGHT]
DigitsDown_1         WORD    0[DIGIT_HEIGHT]
DigitsDown_0         WORD    0[DIGIT_HEIGHT]
DigitsDown_9bot      WORD    0[BOTTOM_DIGIT_ROW_HEIGHT]

' Digits that scroll up
DigitsUp
DigitsUp_1top      WORD    0[TOP_DIGIT_ROW_HEIGHT]
DigitsUp_0         WORD    0[DIGIT_HEIGHT]
DigitsUp_1         WORD    0[DIGIT_HEIGHT]
DigitsUp_2         WORD    0[DIGIT_HEIGHT]
DigitsUp_3         WORD    0[DIGIT_HEIGHT]
DigitsUp_4         WORD    0[DIGIT_HEIGHT]
DigitsUp_5         WORD    0[DIGIT_HEIGHT]
DigitsUp_6         WORD    0[DIGIT_HEIGHT]
DigitsUp_7         WORD    0[DIGIT_HEIGHT]
DigitsUp_8         WORD    0[DIGIT_HEIGHT]
DigitsUp_9         WORD    0[DIGIT_HEIGHT]
DigitsUp_0bot      WORD    0[DIGIT_HEIGHT]
DigitsUp_1bot      WORD    0[BOTTOM_DIGIT_ROW_HEIGHT]

' We are not currently allocating our buffers dynamically.
Rx_Buffer          LONG    0[(RX_BUFFER_SIZE + 3) >> 2]
Tx_Buffer          LONG    0[(TX_BUFFER_SIZE + 3) >> 2]

' The display buffer stores all color layers for all displays.
' Someday it will overlap the PASM objects that are initilaized (loaded into
' cogs) once at the start of operations, to maximize the amount of memory
' available for displays.
'
' One byte of memory stores one color bit layer for two pixels.

DisplayBuffer   LONG 0 [((DISPLAY_COLUMNS * REFRESH_ROWS * COLOR_DEPTH) + 3) >> 2]

' ========================================================================
PUB Main | dcog, hcog, r, c, color, dimcolor, rt
  waitcnt(clkfreq * 4 + cnt)   ' Wait 4 seconds at startup.
  InitFDS

  ' Load the display refresh cog and the write-helper cog, and then tell
  ' the drivers about the two cogs. Note that telling the display driver
  ' about the address of the display buffer will unblock that cog _and_
  ' the helper cog at the same time, so that should be done last.
  dcog := StartDisplayCog
  hcog := StartHelperCog
  Helper.FinalHelperInit(hcog, @DisplayBuffer, COLOR_DEPTH, @CmdFifoData, 320)
  Panel.FinalDisplayInit(dcog, @DisplayBuffer, COLOR_DEPTH)
  ReportDec(string("Started display cog: "), dcog - 1)
  ReportDec(string("Started helper cog: "), hcog - 1)

  'DumpFifo(string("Initial FiFo state"))
  ShowRefreshRate

  ' Capture the ROM Font digits, and prepare the bitmap buffers.
  CaptureAllDigits
  ' Get the display colors
  BoldColorDisplay := Panel.ToDisplayColor(BoldColorRgb)
  DimColorDisplay := Panel.ToDisplayColor(DimColorRgb)
  ReportDec(string("Maximum number of digits across: "), MAX_DIGIT_COLUMNS)
  ReportDec(string("Bits per color: "), COLOR_DEPTH)

  SimpleDigitsReplace(4)
  SlowDigitsReplace(4)
  ScrollDownMultiRow(4)
  ScrollDownMultiRowBoldCenter(4)
  ScrollBothMultiRow(4)
  ScrollBothMultiRowBoldCenter(4)
  if (MAX_DIGIT_COLUMNS => 6)
    SimpleDigitsReplace(6)
    SlowDigitsReplace(6)
    ScrollDownMultiRow(6)
    ScrollDownMultiRowBoldCenter(6)
    ScrollBothMultiRow(6)
    ScrollBothMultiRowBoldCenter(6)
    if MAX_DIGIT_COLUMNS => 8
      SimpleDigitsReplace(8)
      SlowDigitsReplace(8)
      ScrollDownMultiRow(8)
      ScrollDownMultiRowBoldCenter(8)
      ScrollBothMultiRow(8)
      ScrollBothMultiRowBoldCenter(8)

  ' Stall, waiting for input from the user
  ReportStr(string("Press a key to stop the display"))
  Ser.Rx  ' Stall, waiting for a key press, which is discarded.
  Panel.Stop
  ReportStr(string("Display cog stopped"))
  Helper.Stop
  ReportStr(string("Helper cog stopped"))
  repeat  ' wait forvever.

PRI InitFDS | SerialCog
  SerialCog := SerPasm.LoadCog(RXPIN, TXPIN, %0000, BAUDRATE, Ser#NO_CTS, Ser.GetInitVars)
  if SerialCog == 0
    ' Could not start the serial cog -- just hang.
    repeat
      ' Stall
  Ser.FinishInit(serialcog, @Rx_Buffer, RX_BUFFER_SIZE, {
        }                   @Tx_Buffer, TX_BUFFER_SIZE)

PUB StartDisplayCog
  return PanelPasm.LoadDisplayCog(COLOR_DEPTH, Panel.GetDisplayBufAddrVar, {
           } Panel.GetRefreshClocksAddrHub)

PUB StartHelperCog
  return HelperPasm.LoadHelperCog(COLOR_DEPTH, Panel.GetDisplayBufAddrVar, {
           } Helper.GetCmdFifoHeader, @HelperDebugOutput)

PUB ShowRefreshRate | rc
  ' Wait for the first report of the refresh time.
  rc := Panel.GetRefreshClocks
  ReportDec(string("One refresh, approx. sysclks: "), rc)
  ReportTenths(string("Refresh rate, per second: "), {
    }          Umath.multdiv_rounded(CLKFREQ, 10, rc))


PUB CaptureAllDigits
  CaptureDigitPixels("O", @DigitsUp_0) ' Zero has a slash, "O" does not.
  CaptureDigitPixels("1", @DigitsUp_1)
  CaptureDigitPixels("2", @DigitsUp_2)
  CaptureDigitPixels("3", @DigitsUp_3)
  CaptureDigitPixels("4", @DigitsUp_4)
  CaptureDigitPixels("5", @DigitsUp_5)
  CaptureDigitPixels("6", @DigitsUp_6)
  CaptureDigitPixels("7", @DigitsUp_7)
  CaptureDigitPixels("8", @DigitsUp_8)
  CaptureDigitPixels("9", @DigitsUp_9)

  ' Copy them.
  WORDMOVE(@DigitsDown_0, @DigitsUp_0, DIGIT_HEIGHT)
  WORDMOVE(@DigitsDown_1, @DigitsUp_1, DIGIT_HEIGHT)
  WORDMOVE(@DigitsDown_2, @DigitsUp_2, DIGIT_HEIGHT)
  WORDMOVE(@DigitsDown_3, @DigitsUp_3, DIGIT_HEIGHT)
  WORDMOVE(@DigitsDown_4, @DigitsUp_4, DIGIT_HEIGHT)
  WORDMOVE(@DigitsDown_5, @DigitsUp_5, DIGIT_HEIGHT)
  WORDMOVE(@DigitsDown_6, @DigitsUp_6, DIGIT_HEIGHT)
  WORDMOVE(@DigitsDown_7, @DigitsUp_7, DIGIT_HEIGHT)
  WORDMOVE(@DigitsDown_8, @DigitsUp_8, DIGIT_HEIGHT)
  WORDMOVE(@DigitsDown_9, @DigitsUp_9, DIGIT_HEIGHT)

  ' Fill in the ends of the bitmaps with the partial digits
  WORDMOVE(@DigitsUp_0bot  , @DigitsUp_0, DIGIT_HEIGHT)
  WORDMOVE(@DigitsDown_0top, @DigitsUp_0, DIGIT_HEIGHT)
  WORDMOVE(@DigitsUp_1bot  , @DigitsUp_1, BOTTOM_DIGIT_ROW_HEIGHT)
  WORDMOVE(@DigitsDown_9bot, @DigitsUp_9, BOTTOM_DIGIT_ROW_HEIGHT)
  WORDMOVE(@DigitsDown_1top, @DigitsUp_1[DIGIT_HEIGHT - TOP_DIGIT_ROW_HEIGHT], {
        }  TOP_DIGIT_ROW_HEIGHT)
  WORDMOVE(@DigitsUp_1top, @DigitsUp_1[DIGIT_HEIGHT - TOP_DIGIT_ROW_HEIGHT], {
        }  TOP_DIGIT_ROW_HEIGHT)


PUB CaptureDigitPixels(d, p) | vp, v, rdbit, wrbit, o
  ' Capture the pixels from the ROM character font for d in 0..9
  ' into the supplied WORD-based buffer, keeping only the 13 columns of
  ' interest 2..14 (this leaves off the empty columns), and the 26 rows
  ' of interest 3..29 (this leaves off some of the empty rows).

  ' Get the pixels for the pair of characters, skipping over 3 rows
  ' of pixels, which is 3 * 4 == 12 bytes.
  ' In constant terms, we skip over 4 * ((32 - DIGIT_HEIGHT) >> 1) bytes
  vp := $8000 + ((d & $FE) << 6) + (4 * ((32 - DIGIT_HEIGHT) >> 1))
  repeat DIGIT_HEIGHT  ' rows
    v := LONG[vp] >> 4  ' drop 2 columns on the left
    rdbit := 1 << (d & 1)
    wrbit := 1
    o := 0
    repeat DIGIT_WIDTH  ' bits wide
      if (v & rdbit) <> 0
        o |= wrbit
      wrbit <<= 1
      rdbit <<= 2
    WORD[p] := o
    p += 2
    vp += 4

PUB DelaySec(s)
  waitcnt(CNT + (CLKFREQ * s))

' ========================================================================

PUB SimpleDigitsReplace(ndigits) | col, i, startcnt, endcnt
  ' This tests simply writing the center row of digits very fast.
  ReportStr(string("Test: Simple digit replacement"))
  ReportDec(string("Number of digits: "), ndigits)
  Panel.Clear
  ' Show the initial digits
  col := LEFT_DIGIT_LEFT_COLUMN
  DigitBitmapRow[0] := @DigitsUp_0
  DigitValues[0] := 0
  DelaySec(2)

  Helper.Bitmap1Bpp(DigitBitmapRow[0], 2, col, CENTER_DIGIT_TOP_ROW, {
    }               DIGIT_WIDTH, DIGIT_HEIGHT, 0, 0, 0, BoldColorDisplay, 0)
  repeat i from 1 to ndigits - 1
    ' Each bitmap row is a WORD.
    DigitBitmapRow[i] := DigitBitmapRow[i - 1] + (DIGIT_HEIGHT << 1)
    DigitValues[i] := DigitValues[i - 1] + 1
    col += COLUMNS_PER_DIGIT
    Helper.Bitmap1Bpp(DigitBitmapRow[i], 2, col, CENTER_DIGIT_TOP_ROW, {
      }               DIGIT_WIDTH, DIGIT_HEIGHT, 0, 0, 0, BoldColorDisplay, 0)

  ' Wait for display updates to complete.
  Helper.NoOpAndWait
  DelaySec(1)

  ' Record starting CNT
  col := LEFT_DIGIT_LEFT_COLUMN
  startcnt := CNT
  ' Perform the digit updates
  repeat i from 0 to ndigits - 1
    '++DigitValues[i]
    DigitBitmapRow[i] += (DIGIT_HEIGHT << 1)
    Helper.Bitmap1Bpp(DigitBitmapRow[i], 2, col, CENTER_DIGIT_TOP_ROW, {
      }               DIGIT_WIDTH, DIGIT_HEIGHT, 0, 0, 0, BoldColorDisplay, 0)
    col += COLUMNS_PER_DIGIT
    
  ' Wait for display updates to complete.
  Helper.NoOpAndWait
  endcnt := CNT
  ReportDigitRowUpdateSpeed(startcnt, endcnt)
  DelaySec(2)

PUB SlowDigitsReplace(ndigits) | col, startcnt, endcnt, r, c, t
  ' This tests simply writing the center row of digits one pixel row at a time.
  ReportStr(string("Test: Slow digit replacement, one row at a time"))
  ReportDec(string("Number of digits: "), ndigits)
  Panel.Clear
  ' This delay allows the update in about 0.5 seconds (1.9 updates per second)
  t := ((CLKFREQ - 3668592) / DIGIT_HEIGHT) >> 1
  ' Show the initial digits
  col := LEFT_DIGIT_LEFT_COLUMN
  DigitBitmapRow[0] := @DigitsUp_0
  DigitValues[0] := 0
  DelaySec(2)

  Helper.Bitmap1Bpp(DigitBitmapRow[0], 2, col, CENTER_DIGIT_TOP_ROW, {
    }               DIGIT_WIDTH, DIGIT_HEIGHT, 0, 0, 0, BoldColorDisplay, 0)
  repeat c from 1 to ndigits - 1
    ' Each bitmap row is a WORD.
    DigitBitmapRow[c] := DigitBitmapRow[c - 1] + (DIGIT_HEIGHT << 1)
    DigitValues[c] := DigitValues[c - 1] + 1
    col += COLUMNS_PER_DIGIT
    Helper.Bitmap1Bpp(DigitBitmapRow[c], 2, col, CENTER_DIGIT_TOP_ROW, {
      }               DIGIT_WIDTH, DIGIT_HEIGHT, 0, 0, 0, BoldColorDisplay, 0)

  ' Wait for display updates to complete.
  Helper.NoOpAndWait
  DelaySec(1)

  ' Record starting CNT
  col := LEFT_DIGIT_LEFT_COLUMN
  startcnt := CNT
  ' Perform the digit updates
  repeat c from 0 to ndigits - 1
    '++DigitValues[c]
    DigitBitmapRow[c] += (DIGIT_HEIGHT << 1)
  ' Now, draw each row of each digit going across
  repeat r from 0 to constant(DIGIT_HEIGHT - 1)
    col := LEFT_DIGIT_LEFT_COLUMN
    repeat c from 0 to ndigits - 1
      Helper.Bitmap1Bpp(DigitBitmapRow[c], 2, col, CENTER_DIGIT_TOP_ROW + r, {
        }               DIGIT_WIDTH, 1, 0, r, 0, BoldColorDisplay, 0)
      col += COLUMNS_PER_DIGIT
    waitcnt(CNT + t)
    
  ' Wait for display updates to complete.
  Helper.NoOpAndWait
  endcnt := CNT
  ReportDigitRowUpdateSpeed(startcnt, endcnt)
  DelaySec(2)


PUB ScrollDownMultiRow(ndigits) | col, startcnt, endcnt, r, c, t, diff
  ' Digits scroll downward in a loop.
  '
  ' Method: the bitmap row (address) for the top row of the display for each
  ' column is kept, the new row is the calculated, and for each column that
  ' is different, that digit column is updated by one row. This is done up
  ' to 26 times.

  ReportStr(string("Test: Scroll down, multiple digit rows"))
  ReportDec(string("Number of digits: "), ndigits)
  Panel.Clear
  ' This delay allows the update in about 0.5 seconds (1.9 updates per second)
  t := (CLKFREQ / DIGIT_HEIGHT) >> 1
  ' Show the initial digits
  col := LEFT_DIGIT_LEFT_COLUMN
  DigitBitmapRow[0] := @DigitsDown_0  - (TOP_DIGIT_ROW_HEIGHT << 1)
  DigitValues[0] := 0
  DelaySec(2)

  Helper.Bitmap1Bpp(DigitBitmapRow[0], 2, col, 0, {
    }               DIGIT_WIDTH, DISPLAY_ROWS, 0, 0, 0, BoldColorDisplay, 0)
  repeat c from 1 to ndigits - 1
    ' Each bitmap row is a WORD.
    DigitBitmapRow[c] := DigitBitmapRow[c - 1] - (DIGIT_HEIGHT << 1)
    DigitValues[c] := DigitValues[c - 1] + 1
    col += COLUMNS_PER_DIGIT
    Helper.Bitmap1Bpp(DigitBitmapRow[c], 2, col, 0, {
      }               DIGIT_WIDTH, DISPLAY_ROWS, 0, 0, 0, BoldColorDisplay, 0)

  ' Wait for display updates to complete.
  Helper.NoOpAndWait
  DelaySec(1)

  ' Record starting CNT
  col := LEFT_DIGIT_LEFT_COLUMN
  startcnt := CNT
  ' Perform the digit updates
  repeat c from 0 to ndigits - 1
    '++DigitValues[c]
    NewDigitBitmapRow[c] := DigitBitmapRow[c] - (DIGIT_HEIGHT << 1)
  ' Now, draw each row of each digit going across
  diff := 1
  repeat while diff <> 0
    diff := 0
    col := LEFT_DIGIT_LEFT_COLUMN
    repeat c from 0 to ndigits - 1
      if NewDigitBitmapRow[c] <> DigitBitmapRow[c]
        diff := 1
        DigitBitmapRow[c] -= 2  ' back up one row
        ' Should wrap DigitBitmapRow[c]
        Helper.Bitmap1Bpp(DigitBitmapRow[c], 2, col, 0, {
          }               DIGIT_WIDTH, DISPLAY_ROWS, 0, 0, 0, BoldColorDisplay, 0)
      col += COLUMNS_PER_DIGIT
    waitcnt(CNT + t)
    
  ' Wait for display updates to complete.
  Helper.NoOpAndWait
  endcnt := CNT
  ReportDigitRowUpdateSpeed(startcnt, endcnt)
  DelaySec(2)

PUB ScrollDownMultiRowBoldCenter(ndigits) | col, startcnt, endcnt, r, c, t, {
      }                                     diff
  ' Digits scroll downward in a loop. The center rows of the display are bold
  ' (or a different color).
  '
  ' Method: the bitmap row (address) for the top row of the display for each
  ' column is kept, the new row is the calculated, and for each column that
  ' is different, that digit column is updated by one row. This is done up
  ' to 26 times.
  '
  ' Each digit column is divided into three regions: the top and bottom region
  ' are dim, the middle region is bold.
   

  ReportStr(string("Test: Scroll down, bold center digit row"))
  ReportDec(string("Number of digits: "), ndigits)
  Panel.Clear
  ' This delay allows the update in about 0.5 seconds (1.9 updates per second)
  t := (CLKFREQ / DIGIT_HEIGHT) >> 1
  ' Show the initial digits
  col := LEFT_DIGIT_LEFT_COLUMN
  DigitBitmapRow[0] := @DigitsDown_0  - (TOP_DIGIT_ROW_HEIGHT << 1)
  DigitBitmapRowUp[0] := @DigitsUp_0  - (TOP_DIGIT_ROW_HEIGHT << 1)
  DigitValues[0] := 0
  DelaySec(2)

  ' Initial digits display
  Helper.Bitmap1Bpp(DigitBitmapRow[0], 2, col, top_row, {
    }               DIGIT_WIDTH, top_row_height, 0, 0, 0, DimColorDisplay, 0)
  Helper.Bitmap1Bpp(DigitBitmapRow[0] + middle_row_bitmap_offset, 2, col, {
    }               middle_row, DIGIT_WIDTH, middle_row_height, 0, 0, 0, {
    }               BoldColorDisplay, 0)
  Helper.Bitmap1Bpp(DigitBitmapRow[0] + bottom_row_bitmap_offset, 2, col, {
    }               bottom_row, DIGIT_WIDTH, bottom_row_height, 0, 0, 0, {
    }               DimColorDisplay, 0)
  repeat c from 1 to ndigits - 1
    ' Each bitmap row is a WORD.
    DigitBitmapRow[c] := DigitBitmapRow[c - 1] - (DIGIT_HEIGHT << 1)
    DigitValues[c] := DigitValues[c - 1] + 1
    col += COLUMNS_PER_DIGIT
    'Helper.Bitmap1Bpp(DigitBitmapRow[c], 2, col, 0, {
    '  }               DIGIT_WIDTH, DISPLAY_ROWS, 0, 0, 0, BoldColorDisplay, 0)
    Helper.Bitmap1Bpp(DigitBitmapRow[c], 2, col, top_row, {
      }               DIGIT_WIDTH, top_row_height, 0, 0, 0, DimColorDisplay, 0)
    Helper.Bitmap1Bpp(DigitBitmapRow[c] + middle_row_bitmap_offset, 2, col, {
      }               middle_row, DIGIT_WIDTH, middle_row_height, 0, 0, 0, {
      }               BoldColorDisplay, 0)
    Helper.Bitmap1Bpp(DigitBitmapRow[c] + bottom_row_bitmap_offset, 2, col, {
      }               bottom_row, DIGIT_WIDTH, bottom_row_height, 0, 0, 0, {
      }               DimColorDisplay, 0)

  ' Wait for display updates to complete.
  Helper.NoOpAndWait
  DelaySec(1)

  ' Record starting CNT
  col := LEFT_DIGIT_LEFT_COLUMN
  startcnt := CNT
  ' Perform the digit updates
  repeat c from 0 to ndigits - 1
    '++DigitValues[c]
    NewDigitBitmapRow[c] := DigitBitmapRow[c] - (DIGIT_HEIGHT << 1)
  ' Now, draw each row of each digit going across
  diff := 1
  repeat while diff <> 0
    diff := 0
    col := LEFT_DIGIT_LEFT_COLUMN
    repeat c from 0 to ndigits - 1
      if NewDigitBitmapRow[c] <> DigitBitmapRow[c]
        diff := 1
        DigitBitmapRow[c] -= 2  ' back up one row
        ' Should wrap DigitBitmapRow[c]
        'Helper.Bitmap1Bpp(DigitBitmapRow[c], 2, col, 0, {
        '  }               DIGIT_WIDTH, DISPLAY_ROWS, 0, 0, 0, BoldColorDisplay, 0)
        Helper.Bitmap1Bpp(DigitBitmapRow[c], 2, col, top_row, {
          }               DIGIT_WIDTH, top_row_height, 0, 0, 0, DimColorDisplay, 0)
        Helper.Bitmap1Bpp(DigitBitmapRow[c] + middle_row_bitmap_offset, 2, col, {
          }               middle_row, DIGIT_WIDTH, middle_row_height, 0, 0, 0, {
          }               BoldColorDisplay, 0)
        Helper.Bitmap1Bpp(DigitBitmapRow[c] + bottom_row_bitmap_offset, 2, col, {
          }               bottom_row, DIGIT_WIDTH, bottom_row_height, 0, 0, 0, {
          }               DimColorDisplay, 0)
      col += COLUMNS_PER_DIGIT
    waitcnt(CNT + t)
    
  ' Wait for display updates to complete.
  Helper.NoOpAndWait
  endcnt := CNT
  ReportDigitRowUpdateSpeed(startcnt, endcnt)
  DelaySec(2)

PUB ScrollBothMultiRow(ndigits) | col, startcnt, endcnt, r, c, t, diff
  ' Digits scroll in both directions, depending on the pixel column.
  ' THis requires using both the DigitsUp and DigitsDown extended bitmaps.
  '
  ' Method: the bitmap row (address) for the top row of the display for each
  ' column is kept, the new row is the calculated, and for each column that
  ' is different, that digit column is updated by one row. This is done up
  ' to 26 times.

  ReportStr(string("Test: Scroll in two directions, per pixel column, one color"))
  ReportDec(string("Number of digits: "), ndigits)
  Panel.Clear
  ' This delay allows the update in about 0.5 seconds (1.9 updates per second)
  t := (CLKFREQ / DIGIT_HEIGHT) >> 1
  ' Show the initial digits
  col := LEFT_DIGIT_LEFT_COLUMN
  DigitBitmapRow[0] := @DigitsDown_0  - (TOP_DIGIT_ROW_HEIGHT << 1)
  DigitBitmapRowUp[0] := @DigitsUp_0  - (TOP_DIGIT_ROW_HEIGHT << 1)
  DigitValues[0] := 0
  DelaySec(2)

  Helper.Bitmap1Bpp(DigitBitmapRow[0], 2, col, 0, {
    }               DIGIT_WIDTH, DISPLAY_ROWS, 0, 0, 0, BoldColorDisplay, {
    }               SCROLL_DOWN_SKIP_MASK)
  Helper.Bitmap1Bpp(DigitBitmapRowUp[0], 2, col, 0, {
    }               DIGIT_WIDTH, DISPLAY_ROWS, 0, 0, 0, BoldColorDisplay, {
    }               SCROLL_UP_SKIP_MASK)
  repeat c from 1 to ndigits - 1
    ' Each bitmap row is a WORD.
    DigitBitmapRow[c]   := DigitBitmapRow[c - 1] - (DIGIT_HEIGHT << 1)
    DigitBitmapRowUp[c] := DigitBitmapRowUp[c - 1] + (DIGIT_HEIGHT << 1)
    DigitValues[c] := DigitValues[c - 1] + 1
    col += COLUMNS_PER_DIGIT
    Helper.Bitmap1Bpp(DigitBitmapRow[c], 2, col, 0, {
      }               DIGIT_WIDTH, DISPLAY_ROWS, 0, 0, 0, BoldColorDisplay, {
      }               SCROLL_DOWN_SKIP_MASK)
    Helper.Bitmap1Bpp(DigitBitmapRowUp[c], 2, col, 0, {
      }               DIGIT_WIDTH, DISPLAY_ROWS, 0, 0, 0, BoldColorDisplay, {
      }               SCROLL_UP_SKIP_MASK)

  ' Wait for display updates to complete.
  Helper.NoOpAndWait
  DelaySec(1)

  ' Record starting CNT
  col := LEFT_DIGIT_LEFT_COLUMN
  startcnt := CNT
  ' Perform the digit updates
  repeat c from 0 to ndigits - 1
    '++DigitValues[c]
    NewDigitBitmapRow[c] := DigitBitmapRow[c] - (DIGIT_HEIGHT << 1)
  ' Now, draw each row of each digit going across
  diff := 1
  repeat while diff <> 0
    diff := 0
    col := LEFT_DIGIT_LEFT_COLUMN
    repeat c from 0 to ndigits - 1
      if NewDigitBitmapRow[c] <> DigitBitmapRow[c]
        diff := 1
        DigitBitmapRow[c] -= 2  ' back up one row
        DigitBitmapRowUp[c] += 2  ' back up one row
        ' Should wrap DigitBitmapRow[c]
        Helper.Bitmap1Bpp(DigitBitmapRow[c], 2, col, 0, {
          }               DIGIT_WIDTH, DISPLAY_ROWS, 0, 0, 0, BoldColorDisplay, {
          }               SCROLL_DOWN_SKIP_MASK)
        Helper.Bitmap1Bpp(DigitBitmapRowUp[c], 2, col, 0, {
          }               DIGIT_WIDTH, DISPLAY_ROWS, 0, 0, 0, BoldColorDisplay, {
          }               SCROLL_UP_SKIP_MASK)
      col += COLUMNS_PER_DIGIT
    waitcnt(CNT + t)
    
  ' Wait for display updates to complete.
  Helper.NoOpAndWait
  endcnt := CNT
  ReportDigitRowUpdateSpeed(startcnt, endcnt)
  DelaySec(2)

PUB ScrollBothMultiRowBoldCenter(ndigits) | col, startcnt, endcnt, r, c, t, {
      }                                     diff
  ' Digits scroll in both directions, depending on the pixel column.
  ' This requires using both the DigitsUp and DigitsDown extended bitmaps.
  ' The center row of digits is bold (or a different color).
  '
  ' Each digit column is divided into three regions: the top and bottom region
  ' are dim, the middle region is bold.

  ReportStr(string("Test: Scroll in two directions, per pixel column, bold center digit row"))
  ReportDec(string("Number of digits: "), ndigits)
  Panel.Clear
  ' This delay allows the update in about 0.5 seconds (1.9 updates per second)
  t := (CLKFREQ / DIGIT_HEIGHT) >> 1
  ' Show the initial digits
  col := LEFT_DIGIT_LEFT_COLUMN
  DigitBitmapRow[0] := @DigitsDown_0  - (TOP_DIGIT_ROW_HEIGHT << 1)
  DigitBitmapRowUp[0] := @DigitsUp_0  - (TOP_DIGIT_ROW_HEIGHT << 1)
  DigitValues[0] := 0
  DelaySec(2)

  Helper.Bitmap1Bpp(DigitBitmapRow[0], 2, col, top_row, {
    }               DIGIT_WIDTH, top_row_height, 0, 0, 0, DimColorDisplay, {
    }               SCROLL_DOWN_SKIP_MASK)
  Helper.Bitmap1Bpp(DigitBitmapRowUp[0], 2, col, top_row, {
    }               DIGIT_WIDTH, top_row_height, 0, 0, 0, DimColorDisplay, {
    }               SCROLL_UP_SKIP_MASK)
  Helper.Bitmap1Bpp(DigitBitmapRow[0] + middle_row_bitmap_offset, 2, col, {
    }               middle_row, DIGIT_WIDTH, middle_row_height, 0, 0, 0, {
    }               BoldColorDisplay, SCROLL_DOWN_SKIP_MASK)
  Helper.Bitmap1Bpp(DigitBitmapRowUp[0] + middle_row_bitmap_offset, 2, col, {
    }               middle_row, DIGIT_WIDTH, middle_row_height, 0, 0, 0, {
    }               BoldColorDisplay, SCROLL_UP_SKIP_MASK)
  Helper.Bitmap1Bpp(DigitBitmapRow[0] + bottom_row_bitmap_offset, 2, col, {
    }               bottom_row, DIGIT_WIDTH, bottom_row_height, 0, 0, 0, {
    }               DimColorDisplay, SCROLL_DOWN_SKIP_MASK)
  Helper.Bitmap1Bpp(DigitBitmapRowUp[0] + bottom_row_bitmap_offset, 2, col, {
    }               bottom_row, DIGIT_WIDTH, bottom_row_height, 0, 0, 0, {
    }               DimColorDisplay, SCROLL_UP_SKIP_MASK)


  repeat c from 1 to ndigits - 1
    ' Each bitmap row is a WORD.
    DigitBitmapRow[c]   := DigitBitmapRow[c - 1] - (DIGIT_HEIGHT << 1)
    DigitBitmapRowUp[c] := DigitBitmapRowUp[c - 1] + (DIGIT_HEIGHT << 1)
    DigitValues[c] := DigitValues[c - 1] + 1
    col += COLUMNS_PER_DIGIT
    'Helper.Bitmap1Bpp(DigitBitmapRow[c], 2, col, 0, {
    '  }               DIGIT_WIDTH, DISPLAY_ROWS, 0, 0, 0, BoldColorDisplay, {
    '  }               SCROLL_DOWN_SKIP_MASK)
    'Helper.Bitmap1Bpp(DigitBitmapRowUp[c], 2, col, 0, {
    '  }               DIGIT_WIDTH, DISPLAY_ROWS, 0, 0, 0, BoldColorDisplay, {
    '  }               SCROLL_UP_SKIP_MASK)
    Helper.Bitmap1Bpp(DigitBitmapRow[c], 2, col, top_row, {
      }               DIGIT_WIDTH, top_row_height, 0, 0, 0, DimColorDisplay, {
      }               SCROLL_DOWN_SKIP_MASK)
    Helper.Bitmap1Bpp(DigitBitmapRowUp[c], 2, col, top_row, {
      }               DIGIT_WIDTH, top_row_height, 0, 0, 0, DimColorDisplay, {
      }               SCROLL_UP_SKIP_MASK)
    Helper.Bitmap1Bpp(DigitBitmapRow[c] + middle_row_bitmap_offset, 2, col, {
      }               middle_row, DIGIT_WIDTH, middle_row_height, 0, 0, 0, {
      }               BoldColorDisplay, SCROLL_DOWN_SKIP_MASK)
    Helper.Bitmap1Bpp(DigitBitmapRowUp[c] + middle_row_bitmap_offset, 2, col, {
      }               middle_row, DIGIT_WIDTH, middle_row_height, 0, 0, 0, {
      }               BoldColorDisplay, SCROLL_UP_SKIP_MASK)
    Helper.Bitmap1Bpp(DigitBitmapRow[c] + bottom_row_bitmap_offset, 2, col, {
      }               bottom_row, DIGIT_WIDTH, bottom_row_height, 0, 0, 0, {
      }               DimColorDisplay, SCROLL_DOWN_SKIP_MASK)
    Helper.Bitmap1Bpp(DigitBitmapRowUp[c] + bottom_row_bitmap_offset, 2, col, {
      }               bottom_row, DIGIT_WIDTH, bottom_row_height, 0, 0, 0, {
      }               DimColorDisplay, SCROLL_UP_SKIP_MASK)

  ' Wait for display updates to complete.
  Helper.NoOpAndWait
  DelaySec(1)

  ' Record starting CNT
  col := LEFT_DIGIT_LEFT_COLUMN
  startcnt := CNT
  ' Perform the digit updates
  repeat c from 0 to ndigits - 1
    '++DigitValues[c]
    NewDigitBitmapRow[c] := DigitBitmapRow[c] - (DIGIT_HEIGHT << 1)
  ' Now, draw each row of each digit going across
  diff := 1
  repeat while diff <> 0
    diff := 0
    col := LEFT_DIGIT_LEFT_COLUMN
    repeat c from 0 to ndigits - 1
      if NewDigitBitmapRow[c] <> DigitBitmapRow[c]
        diff := 1
        DigitBitmapRow[c] -= 2  ' back up one row
        DigitBitmapRowUp[c] += 2  ' back up one row
        ' Should wrap DigitBitmapRow[c]
        'Helper.Bitmap1Bpp(DigitBitmapRow[c], 2, col, 0, {
        '  }               DIGIT_WIDTH, DISPLAY_ROWS, 0, 0, 0, BoldColorDisplay, {
        '  }               SCROLL_DOWN_SKIP_MASK)
        'Helper.Bitmap1Bpp(DigitBitmapRowUp[c], 2, col, 0, {
        '  }               DIGIT_WIDTH, DISPLAY_ROWS, 0, 0, 0, BoldColorDisplay, {
        '  }               SCROLL_UP_SKIP_MASK)
        Helper.Bitmap1Bpp(DigitBitmapRow[c], 2, col, top_row, {
          }               DIGIT_WIDTH, top_row_height, 0, 0, 0, DimColorDisplay, {
          }               SCROLL_DOWN_SKIP_MASK)
        Helper.Bitmap1Bpp(DigitBitmapRowUp[c], 2, col, top_row, {
          }               DIGIT_WIDTH, top_row_height, 0, 0, 0, DimColorDisplay, {
          }               SCROLL_UP_SKIP_MASK)
        Helper.Bitmap1Bpp(DigitBitmapRow[c] + middle_row_bitmap_offset, 2, col, {
          }               middle_row, DIGIT_WIDTH, middle_row_height, 0, 0, 0, {
          }               BoldColorDisplay, SCROLL_DOWN_SKIP_MASK)
        Helper.Bitmap1Bpp(DigitBitmapRowUp[c] + middle_row_bitmap_offset, 2, col, {
          }               middle_row, DIGIT_WIDTH, middle_row_height, 0, 0, 0, {
          }               BoldColorDisplay, SCROLL_UP_SKIP_MASK)
        Helper.Bitmap1Bpp(DigitBitmapRow[c] + bottom_row_bitmap_offset, 2, col, {
          }               bottom_row, DIGIT_WIDTH, bottom_row_height, 0, 0, 0, {
          }               DimColorDisplay, SCROLL_DOWN_SKIP_MASK)
        Helper.Bitmap1Bpp(DigitBitmapRowUp[c] + bottom_row_bitmap_offset, 2, col, {
          }               bottom_row, DIGIT_WIDTH, bottom_row_height, 0, 0, 0, {
          }               DimColorDisplay, SCROLL_UP_SKIP_MASK)
      col += COLUMNS_PER_DIGIT
    waitcnt(CNT + t)
    
  ' Wait for display updates to complete.
  Helper.NoOpAndWait
  endcnt := CNT
  ReportDigitRowUpdateSpeed(startcnt, endcnt)
  DelaySec(2)

' ========================================================================

PUB ReportStr(s)
  Ser.Str(s)
  Ser.Tx(CR)

PUB ReportDec(label, v)
  Ser.Str(label)
  Ser.Dec(v)
  Ser.Tx(CR)

PUB ReportHex(label, v)
  Ser.Str(label)
  Ser.Hex(v, 8)
  Ser.Tx(CR)

PUB ReportBin(label, v)
  Ser.Str(label)
  Ser.Bin(v, 24)
  Ser.Tx(CR)

Pub ReportTenths(label, v)
  Ser.Str(label)
  Ser.Dec(v / 10)
  Ser.Tx(".")
  Ser.Tx("0" + (v // 10))
  Ser.Tx(CR)

PUB DumpFifo(label) | p
  Ser.Str(label)
  Ser.Tx(CR)
  p := Helper.GetCmdFifoHeader
  
  ReportHex(string("FiFo   ptr: "), p)
  ReportHex(string("Insert ptr: "), word[p + Fifo#INSERT_HUBADDR_OFFSET])
  ReportHex(string("Remove ptr: "), word[p + Fifo#REMOVE_HUBADDR_OFFSET])
  ReportHex(string("Start  ptr: "), word[p + Fifo#START_HUBADDR_OFFSET])
  ReportHex(string("Limit  ptr: "), word[p + Fifo#LIMIT_HUBADDR_OFFSET])
  ReportDec(string("Entry size: "), word[p + Fifo#ENTRY_SIZE_BYTES_OFFSET])

PUB ReportBitmapDebugOutput(entry)
  ReportHex(string("result address:  "), entry)
  ReportHex(string("Requested background color: "), long[entry + 20])
  ReportHex(string("Seen      background color: "), long[entry +  8])
  ReportBin(string("Converted background color: "), long[entry +  0])
  ReportHex(string("Requested foreground color: "), long[entry + 24])
  ReportHex(string("Seen      foreground color: "), long[entry + 12])
  ReportBin(string("Converted foreground color: "), long[entry +  4])

PUB ReportRomCharDebugOutput(entry)
  ReportHex(string("result address:  "), entry)
  ReportHex(string("Rom char base address: "), long[entry +  0])
  ReportHex(string("Rom char seen:         "), long[entry +  4])
  ReportHex(string("Rom char requested:    "), word[entry + 16])
  ReportHex(string("Expected base address: "), ((word[entry + 16] & $FE) << 6) + $8000)

PUB ReportSetColorDebugOutput(entry) | c
  c := long[entry + 8]
  ReportHex(string("Original      color value: "), c)
  ReportBin(string("Converted raw color value: "), long[entry + 4])
  ReportBin(string("Expected  raw color value: "), {
    } Panel.DisplayColor(((c >> 16) & $FF), ((c >> 8) & $FF), c & $FF))
  'ReportBin(string("Expected2 raw color value: "), Panel.DisplayColor(6, 6, 0))

PUB ReportDigitRowUpdateSpeed(startcnt, endcnt) | clks
  clks := endcnt - startcnt
  ReportDec(string("Sysclks to update all digits: "), clks)
  ' calculate and report update rate, to tenths.
  ReportTenths(string("Update rate, per second: "), {
    }          Umath.multdiv_rounded(CLKFREQ, 10, clks))
  'ReportDec(string("Update rate, in tenths: "), {
  '  }          Umath.multdiv_rounded(CLKFREQ, 10, clks))
  
